# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import re

import pytest

from playwright.sync_api import Browser, Error, Page, expect
from tests.server import Server


def test_assertions_page_to_have_title(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<title>new title</title>")
    expect(page).to_have_title("new title")
    expect(page).to_have_title(re.compile("new title"))
    with pytest.raises(AssertionError):
        expect(page).to_have_title("not the current title", timeout=750)
    with pytest.raises(AssertionError):
        expect(page).to_have_title(re.compile("not the current title"), timeout=750)
    with pytest.raises(AssertionError):
        expect(page).not_to_have_title(re.compile("new title"), timeout=750)
    with pytest.raises(AssertionError):
        expect(page).not_to_have_title("new title", timeout=750)
    expect(page).not_to_have_title("great title", timeout=750)
    page.evaluate(
        """
        setTimeout(() => {
            document.title = 'great title';
        }, 2000);
    """
    )
    expect(page).to_have_title("great title")
    expect(page).to_have_title(re.compile("great title"))


def test_assertions_page_to_have_url(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    expect(page).to_have_url(server.EMPTY_PAGE)
    expect(page).to_have_url(re.compile(r".*/empty\.html"))
    with pytest.raises(AssertionError):
        expect(page).to_have_url("nooooo", timeout=750)
    with pytest.raises(AssertionError):
        expect(page).to_have_url(re.compile("not-the-url"), timeout=750)
    page.evaluate(
        """
        setTimeout(() => {
            window.location = window.location.origin + '/grid.html';
        }, 2000);
    """
    )
    expect(page).to_have_url(server.PREFIX + "/grid.html")
    expect(page).not_to_have_url(server.EMPTY_PAGE, timeout=750)
    with pytest.raises(AssertionError):
        expect(page).not_to_have_url(re.compile(r".*/grid\.html"), timeout=750)
    with pytest.raises(AssertionError):
        expect(page).not_to_have_url(server.PREFIX + "/grid.html", timeout=750)
    expect(page).to_have_url(re.compile(r".*/grid\.html"))
    expect(page).not_to_have_url("**/empty.html", timeout=750)


def test_assertions_page_to_have_url_with_base_url(
    browser: Browser, server: Server
) -> None:
    page = browser.new_page(base_url=server.PREFIX)
    page.goto("/empty.html")
    expect(page).to_have_url("/empty.html")
    expect(page).to_have_url(re.compile(r".*/empty\.html"))
    page.close()


def test_assertions_locator_to_contain_text(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div id=foobar>kek</div>")
    expect(page.locator("div#foobar")).to_contain_text("kek")
    expect(page.locator("div#foobar")).not_to_contain_text("bar", timeout=100)
    with pytest.raises(AssertionError):
        expect(page.locator("div#foobar")).to_contain_text("bar", timeout=100)

    page.set_content("<div>Text \n1</div><div>Text2</div><div>Text3</div>")
    expect(page.locator("div")).to_contain_text(["ext     1", re.compile("ext3")])


def test_assertions_locator_to_have_attribute(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div id=foobar>kek</div>")
    expect(page.locator("div#foobar")).to_have_attribute("id", "foobar")
    expect(page.locator("div#foobar")).to_have_attribute("id", re.compile("foobar"))
    expect(page.locator("div#foobar")).not_to_have_attribute("id", "kek", timeout=100)
    with pytest.raises(AssertionError):
        expect(page.locator("div#foobar")).to_have_attribute("id", "koko", timeout=100)


def test_assertions_locator_to_have_attribute_ignore_case(
    page: Page, server: Page
) -> None:
    page.set_content("<div id=NoDe>Text content</div>")
    locator = page.locator("#NoDe")
    expect(locator).to_have_attribute("id", "node", ignore_case=True)
    expect(locator).not_to_have_attribute("id", "node")


def test_assertions_locator_to_have_class(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div class=foobar>kek</div>")
    expect(page.locator("div.foobar")).to_have_class("foobar")
    expect(page.locator("div.foobar")).to_have_class(["foobar"])
    expect(page.locator("div.foobar")).to_have_class(re.compile("foobar"))
    expect(page.locator("div.foobar")).to_have_class([re.compile("foobar")])
    expect(page.locator("div.foobar")).not_to_have_class("kekstar", timeout=100)
    with pytest.raises(AssertionError):
        expect(page.locator("div.foobar")).to_have_class("oh-no", timeout=100)


def test_assertions_locator_to_contain_class(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div class='foo bar baz'></div>")
    locator = page.locator("div")
    expect(locator).to_contain_class("")
    expect(locator).to_contain_class("bar")
    expect(locator).to_contain_class("baz bar")
    expect(locator).to_contain_class("  bar   foo ")
    expect(locator).not_to_contain_class(
        "  baz   not-matching "
    )  # Strip whitespace and match individual classes
    with pytest.raises(AssertionError) as excinfo:
        expect(locator).to_contain_class("does-not-exist", timeout=100)

    assert excinfo.match("Locator expected to contain class 'does-not-exist'")
    assert excinfo.match("Actual value: foo bar baz")
    assert excinfo.match('Expect "to_contain_class" with timeout 100ms')

    page.set_content(
        '<div class="foo"></div><div class="hello bar"></div><div class="baz"></div>'
    )
    expect(locator).to_contain_class(["foo", "hello", "baz"])
    expect(locator).not_to_contain_class(["not-there", "hello", "baz"])
    expect(locator).not_to_contain_class(["foo", "hello"])


def test_assertions_locator_to_have_count(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div class=foobar>kek</div><div class=foobar>kek</div>")
    expect(page.locator("div.foobar")).to_have_count(2)
    expect(page.locator("div.foobar")).not_to_have_count(42, timeout=100)
    with pytest.raises(AssertionError):
        expect(page.locator("div.foobar")).to_have_count(42, timeout=100)


def test_assertions_locator_to_have_css(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div class=foobar style='color: rgb(234, 74, 90);'>kek</div>")
    expect(page.locator("div.foobar")).to_have_css("color", "rgb(234, 74, 90)")
    expect(page.locator("div.foobar")).not_to_have_css(
        "color", "rgb(42, 42, 42)", timeout=100
    )
    with pytest.raises(AssertionError):
        expect(page.locator("div.foobar")).to_have_css(
            "color", "rgb(42, 42, 42)", timeout=100
        )


def test_assertions_locator_to_have_id(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div class=foobar id=kek>kek</div>")
    expect(page.locator("div.foobar")).to_have_id("kek")
    expect(page.locator("div.foobar")).not_to_have_id("top", timeout=100)
    with pytest.raises(AssertionError):
        expect(page.locator("div.foobar")).to_have_id("top", timeout=100)


def test_assertions_locator_to_have_js_property(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div></div>")
    page.eval_on_selector(
        "div", "e => e.foo = { a: 1, b: 'string', c: new Date(1627503992000) }"
    )
    expect(page.locator("div")).to_have_js_property(
        "foo",
        {
            "a": 1,
            "b": "string",
            "c": datetime.datetime.fromtimestamp(1627503992000 / 1000),
        },
    )


def test_to_have_js_property_pass_string(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = 'string'")
    locator = page.locator("div")
    expect(locator).to_have_js_property("foo", "string")


def test_to_have_js_property_fail_string(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = 'string'")
    locator = page.locator("div")
    with pytest.raises(AssertionError):
        expect(locator).to_have_js_property("foo", "error", timeout=500)


def test_to_have_js_property_pass_number(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = 2021")
    locator = page.locator("div")
    expect(locator).to_have_js_property("foo", 2021)


def test_to_have_js_property_fail_number(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = 2021")
    locator = page.locator("div")
    with pytest.raises(AssertionError):
        expect(locator).to_have_js_property("foo", 1, timeout=500)


def test_to_have_js_property_pass_boolean(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = true")
    locator = page.locator("div")
    expect(locator).to_have_js_property("foo", True)


def test_to_have_js_property_fail_boolean(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = false")
    locator = page.locator("div")
    with pytest.raises(AssertionError):
        expect(locator).to_have_js_property("foo", True, timeout=500)


def test_to_have_js_property_pass_boolean_2(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = false")
    locator = page.locator("div")
    expect(locator).to_have_js_property("foo", False)


def test_to_have_js_property_fail_boolean_2(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = false")
    locator = page.locator("div")
    with pytest.raises(AssertionError):
        expect(locator).to_have_js_property("foo", True, timeout=500)


def test_to_have_js_property_pass_null(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector("div", "e => e.foo = null")
    locator = page.locator("div")
    expect(locator).to_have_js_property("foo", None)


def test_assertions_locator_to_have_text(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div id=foobar>kek</div>")
    expect(page.locator("div#foobar")).to_have_text("kek")
    expect(page.locator("div#foobar")).not_to_have_text("top", timeout=100)

    page.set_content("<div>Text    \n1</div><div>Text   2a</div>")
    # Should only normalize whitespace in the first item.
    expect(page.locator("div")).to_have_text(["Text  1", re.compile(r"Text   \d+a")])


@pytest.mark.parametrize(
    "method",
    ["to_have_text", "to_contain_text"],
)
def test_ignore_case(page: Page, server: Server, method: str) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div id=target>apple BANANA</div><div>orange</div>")
    getattr(expect(page.locator("div#target")), method)("apple BANANA")
    getattr(expect(page.locator("div#target")), method)(
        "apple banana", ignore_case=True
    )
    # defaults false
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div#target")), method)(
            "apple banana",
            timeout=300,
        )
    expected_error_msg = method.replace("_", " ")
    assert expected_error_msg in str(excinfo.value)

    # Array Variants
    getattr(expect(page.locator("div")), method)(["apple BANANA", "orange"])
    getattr(expect(page.locator("div")), method)(
        ["apple banana", "ORANGE"], ignore_case=True
    )
    # defaults false
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div")), method)(
            ["apple banana", "ORANGE"],
            timeout=300,
        )
    assert expected_error_msg in str(excinfo.value)

    # not variant
    getattr(expect(page.locator("div#target")), f"not_{method}")("apple banana")
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div#target")), f"not_{method}")(
            "apple banana",
            ignore_case=True,
            timeout=300,
        )
    assert f"not {expected_error_msg}" in str(excinfo)


@pytest.mark.parametrize(
    "method",
    ["to_have_text", "to_contain_text"],
)
def test_ignore_case_regex(page: Page, server: Server, method: str) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div id=target>apple BANANA</div><div>orange</div>")
    getattr(expect(page.locator("div#target")), method)(re.compile("apple BANANA"))
    getattr(expect(page.locator("div#target")), method)(
        re.compile("apple banana"), ignore_case=True
    )
    # defaults to regex flag
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div#target")), method)(
            re.compile("apple banana"), timeout=300
        )
    expected_error_msg = method.replace("_", " ")
    assert expected_error_msg in str(excinfo.value)
    # overrides regex flag
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div#target")), method)(
            re.compile("apple banana", re.IGNORECASE),
            ignore_case=False,
            timeout=300,
        )
    assert expected_error_msg in str(excinfo.value)

    # Array Variants
    getattr(expect(page.locator("div")), method)(
        [re.compile("apple BANANA"), re.compile("orange")]
    )
    getattr(expect(page.locator("div")), method)(
        [re.compile("apple banana"), re.compile("ORANGE")], ignore_case=True
    )
    # defaults regex flag
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div")), method)(
            [re.compile("apple banana"), re.compile("ORANGE")],
            timeout=300,
        )
    assert expected_error_msg in str(excinfo.value)
    # overrides regex flag
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div")), method)(
            [
                re.compile("apple banana", re.IGNORECASE),
                re.compile("ORANGE", re.IGNORECASE),
            ],
            ignore_case=False,
            timeout=300,
        )
    assert expected_error_msg in str(excinfo.value)

    # not variant
    getattr(expect(page.locator("div#target")), f"not_{method}")(
        re.compile("apple banana")
    )
    with pytest.raises(AssertionError) as excinfo:
        getattr(expect(page.locator("div#target")), f"not_{method}")(
            re.compile("apple banana"),
            ignore_case=True,
            timeout=300,
        )
    assert f"not {expected_error_msg}" in str(excinfo)


def test_assertions_locator_to_have_value(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<input type=text id=foo>")
    my_input = page.locator("#foo")
    expect(my_input).to_have_value("")
    expect(my_input).not_to_have_value("bar", timeout=100)
    my_input.fill("kektus")
    expect(my_input).to_have_value("kektus")


def test_to_have_values_works_with_text(page: Page, server: Server) -> None:
    page.set_content(
        """
        <select multiple>
            <option value="R">Red</option>
            <option value="G">Green</option>
            <option value="B">Blue</option>
        </select>
    """
    )
    locator = page.locator("select")
    locator.select_option(["R", "G"])
    expect(locator).to_have_values(["R", "G"])


def test_to_have_values_follows_labels(page: Page, server: Server) -> None:
    page.set_content(
        """
        <label for="colors">Pick a Color</label>
        <select id="colors" multiple>
            <option value="R">Red</option>
            <option value="G">Green</option>
            <option value="B">Blue</option>
        </select>
    """
    )
    locator = page.locator("text=Pick a Color")
    locator.select_option(["R", "G"])
    expect(locator).to_have_values(["R", "G"])


def test_to_have_values_exact_match_with_text(page: Page, server: Server) -> None:
    page.set_content(
        """
        <select multiple>
            <option value="RR">Red</option>
            <option value="GG">Green</option>
        </select>
    """
    )
    locator = page.locator("select")
    locator.select_option(["RR", "GG"])
    with pytest.raises(AssertionError) as excinfo:
        expect(locator).to_have_values(["R", "G"], timeout=500)
    assert "Locator expected to have Values '['R', 'G']'" in str(excinfo.value)
    assert "Actual value: ['RR', 'GG']" in str(excinfo.value)


def test_to_have_values_works_with_regex(page: Page, server: Server) -> None:
    page.set_content(
        """
        <select multiple>
            <option value="R">Red</option>
            <option value="G">Green</option>
            <option value="B">Blue</option>
        </select>
    """
    )
    locator = page.locator("select")
    locator.select_option(["R", "G"])
    expect(locator).to_have_values([re.compile("R"), re.compile("G")])


def test_to_have_values_fails_when_items_not_selected(
    page: Page, server: Server
) -> None:
    page.set_content(
        """
        <select multiple>
            <option value="R">Red</option>
            <option value="G">Green</option>
            <option value="B">Blue</option>
        </select>
    """
    )
    locator = page.locator("select")
    locator.select_option(["B"])
    with pytest.raises(AssertionError) as excinfo:
        expect(locator).to_have_values(["R", "G"], timeout=500)
    assert "Locator expected to have Values '['R', 'G']'" in str(excinfo.value)
    assert "Actual value: ['B']" in str(excinfo.value)


def test_to_have_values_fails_when_multiple_not_specified(
    page: Page, server: Server
) -> None:
    page.set_content(
        """
        <select>
            <option value="R">Red</option>
            <option value="G">Green</option>
            <option value="B">Blue</option>
        </select>
    """
    )
    locator = page.locator("select")
    locator.select_option(["B"])
    with pytest.raises(Error) as excinfo:
        expect(locator).to_have_values(["R", "G"], timeout=500)
    assert "Error: Not a select element with a multiple attribute" in str(excinfo.value)


def test_to_have_values_fails_when_not_a_select_element(
    page: Page, server: Server
) -> None:
    page.set_content(
        """
        <input type="text">
    """
    )
    locator = page.locator("input")
    with pytest.raises(Error) as excinfo:
        expect(locator).to_have_values(["R", "G"], timeout=500)
    assert "Error: Not a select element with a multiple attribute" in str(excinfo.value)


def test_assertions_locator_to_be_checked(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<input type=checkbox>")
    my_checkbox = page.locator("input")
    expect(my_checkbox).not_to_be_checked()
    with pytest.raises(AssertionError, match="Locator expected to be checked"):
        expect(my_checkbox).to_be_checked(timeout=100)
    expect(my_checkbox).to_be_checked(timeout=100, checked=False)
    with pytest.raises(AssertionError):
        expect(my_checkbox).to_be_checked(timeout=100, checked=True)
    my_checkbox.check()
    expect(my_checkbox).to_be_checked(timeout=100, checked=True)
    with pytest.raises(AssertionError, match="Locator expected to be unchecked"):
        expect(my_checkbox).to_be_checked(timeout=100, checked=False)
    expect(my_checkbox).to_be_checked()


def test_assertions_boolean_checked_with_intermediate_true(page: Page) -> None:
    page.set_content("<input type=checkbox></input>")
    page.locator("input").evaluate("e => e.indeterminate = true")
    expect(page.locator("input")).to_be_checked(indeterminate=True)


def test_assertions_boolean_checked_with_intermediate_true_and_checked(
    page: Page,
) -> None:
    page.set_content("<input type=checkbox></input>")
    page.locator("input").evaluate("e => e.indeterminate = true")
    with pytest.raises(
        Error, match="Can't assert indeterminate and checked at the same time"
    ):
        expect(page.locator("input")).to_be_checked(checked=False, indeterminate=True)


def test_assertions_boolean_fail_with_indeterminate_true(page: Page) -> None:
    page.set_content("<input type=checkbox></input>")
    with pytest.raises(
        AssertionError, match='Expect "to_be_checked" with timeout 1000ms'
    ):
        expect(page.locator("input")).to_be_checked(indeterminate=True, timeout=1000)


def test_assertions_locator_to_be_disabled_enabled(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<input type=checkbox>")
    my_checkbox = page.locator("input")
    expect(my_checkbox).not_to_be_disabled()
    expect(my_checkbox).to_be_enabled()
    with pytest.raises(AssertionError):
        expect(my_checkbox).to_be_disabled(timeout=100)
    my_checkbox.evaluate("e => e.disabled = true")
    expect(my_checkbox).to_be_disabled()
    with pytest.raises(AssertionError, match="Locator expected to be enabled"):
        expect(my_checkbox).to_be_enabled(timeout=100)


def test_assertions_locator_to_be_enabled_with_true(page: Page) -> None:
    page.set_content("<button>Text</button>")
    expect(page.locator("button")).to_be_enabled(enabled=True)


def test_assertions_locator_to_be_enabled_with_false_throws_good_exception(
    page: Page,
) -> None:
    page.set_content("<button>Text</button>")
    with pytest.raises(AssertionError, match="Locator expected to be disabled"):
        expect(page.locator("button")).to_be_enabled(enabled=False)


def test_assertions_locator_to_be_enabled_with_false(page: Page) -> None:
    page.set_content("<button disabled>Text</button>")
    expect(page.locator("button")).to_be_enabled(enabled=False)


def test_assertions_locator_to_be_enabled_with_not_and_false(page: Page) -> None:
    page.set_content("<button>Text</button>")
    expect(page.locator("button")).not_to_be_enabled(enabled=False)


def test_assertions_locator_to_be_enabled_eventually(page: Page) -> None:
    page.set_content("<button disabled>Text</button>")
    page.eval_on_selector(
        "button",
        """
        button => setTimeout(() => {
            button.removeAttribute('disabled');
        }, 700);
    """,
    )
    expect(page.locator("button")).to_be_enabled()


def test_assertions_locator_to_be_enabled_eventually_with_not(page: Page) -> None:
    page.set_content("<button>Text</button>")
    page.eval_on_selector(
        "button",
        """
        button => setTimeout(() => {
            button.setAttribute('disabled', '');
        }, 700);
    """,
    )
    expect(page.locator("button")).not_to_be_enabled()


def test_assertions_locator_to_be_editable(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<input></input>")
    expect(page.locator("input")).to_be_editable()


def test_assertions_locator_to_be_editable_throws(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<button disabled>Text</button>")
    with pytest.raises(
        Error,
        match=r"Element is not an <input>, <textarea>, <select> or \[contenteditable\] and does not have a role allowing \[aria-readonly\]",
    ):
        expect(page.locator("button")).not_to_be_editable()


def test_assertions_locator_to_be_editable_with_true(page: Page) -> None:
    page.set_content("<input></input>")
    expect(page.locator("input")).to_be_editable(editable=True)


def test_assertions_locator_to_be_editable_with_false(page: Page) -> None:
    page.set_content("<input readonly></input>")
    expect(page.locator("input")).to_be_editable(editable=False)


def test_assertions_locator_to_be_editable_with_false_and_throw_good_exception(
    page: Page,
) -> None:
    page.set_content("<input></input>")
    with pytest.raises(AssertionError, match="Locator expected to be readonly"):
        expect(page.locator("input")).to_be_editable(editable=False)


def test_assertions_locator_to_be_editable_with_not_and_false(page: Page) -> None:
    page.set_content("<input></input>")
    expect(page.locator("input")).not_to_be_editable(editable=False)


def test_assertions_locator_to_be_empty(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content(
        "<input value=text name=input1></input><input name=input2></input>"
    )
    expect(page.locator("input[name=input1]")).not_to_be_empty()
    expect(page.locator("input[name=input2]")).to_be_empty()
    with pytest.raises(AssertionError):
        expect(page.locator("input[name=input1]")).to_be_empty(timeout=100)


def test_assertions_locator_to_be_focused(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<input type=checkbox>")
    my_checkbox = page.locator("input")
    with pytest.raises(AssertionError):
        expect(my_checkbox).to_be_focused(timeout=100)
    my_checkbox.focus()
    expect(my_checkbox).to_be_focused()


def test_assertions_locator_to_be_hidden_visible(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div style='width: 50px; height: 50px;'>Something</div>")
    my_checkbox = page.locator("div")
    expect(my_checkbox).to_be_visible()
    with pytest.raises(AssertionError):
        expect(my_checkbox).to_be_hidden(timeout=100)
    my_checkbox.evaluate("e => e.style.display = 'none'")
    expect(my_checkbox).to_be_hidden()
    with pytest.raises(AssertionError, match="Locator expected to be visible"):
        expect(my_checkbox).to_be_visible(timeout=100)


def test_assertions_locator_to_be_visible_with_true(page: Page) -> None:
    page.set_content("<button>hello</button>")
    expect(page.locator("button")).to_be_visible(visible=True)


def test_assertions_locator_to_be_visible_with_false(page: Page) -> None:
    page.set_content("<button hidden>hello</button>")
    expect(page.locator("button")).to_be_visible(visible=False)


def test_assertions_locator_to_be_visible_with_false_throws_good_exception(
    page: Page,
) -> None:
    page.set_content("<button>hello</button>")
    with pytest.raises(AssertionError, match="Locator expected to be hidden"):
        expect(page.locator("button")).to_be_visible(visible=False)


def test_assertions_locator_to_be_visible_with_not_and_false(page: Page) -> None:
    page.set_content("<button>hello</button>")
    expect(page.locator("button")).not_to_be_visible(visible=False)


def test_assertions_locator_to_be_visible_eventually(page: Page) -> None:
    page.set_content("<div></div>")
    page.eval_on_selector(
        "div",
        """
        div => setTimeout(() => {
            div.innerHTML = '<span>Hello</span>';
        }, 700);
    """,
    )
    expect(page.locator("span")).to_be_visible()


def test_assertions_locator_to_be_visible_eventually_with_not(page: Page) -> None:
    page.set_content("<div><span>Hello</span></div>")
    page.eval_on_selector(
        "span",
        """
        span => setTimeout(() => {
            span.textContent = '';
        }, 700);
    """,
    )
    expect(page.locator("span")).not_to_be_visible()


def test_assertions_should_serialize_regexp_correctly(
    page: Page, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<div>iGnOrEcAsE</div>")
    expect(page.locator("div")).to_have_text(re.compile(r"ignorecase", re.IGNORECASE))
    page.set_content(
        """<div>start
some
lines
between
end</div>"""
    )
    expect(page.locator("div")).to_have_text(re.compile(r"start.*end", re.DOTALL))
    page.set_content(
        """<div>line1
line2
line3</div>"""
    )
    expect(page.locator("div")).to_have_text(re.compile(r"^line2$", re.MULTILINE))


def test_assertions_response_is_ok_pass(page: Page, server: Server) -> None:
    response = page.request.get(server.EMPTY_PAGE)
    expect(response).to_be_ok()


def test_assertions_response_is_ok_pass_with_not(page: Page, server: Server) -> None:
    response = page.request.get(server.PREFIX + "/unknown")
    expect(response).not_to_be_ok()


def test_assertions_response_is_ok_fail(page: Page, server: Server) -> None:
    response = page.request.get(server.PREFIX + "/unknown")
    with pytest.raises(AssertionError) as excinfo:
        expect(response).to_be_ok()
    error_message = str(excinfo.value)
    assert ("→ GET " + server.PREFIX + "/unknown") in error_message
    assert "← 404 Not Found" in error_message


def test_should_print_response_with_text_content_type_if_to_be_ok_fails(
    page: Page, server: Server
) -> None:
    server.set_route(
        "/text-content-type",
        lambda r: (
            r.setResponseCode(404),
            r.setHeader("content-type", "text/plain"),
            r.write(b"Text error"),
            r.finish(),
        ),
    )
    server.set_route(
        "/no-content-type",
        lambda r: (
            r.setResponseCode(404),
            r.write(b"No content type error"),
            r.finish(),
        ),
    )
    server.set_route(
        "/binary-content-type",
        lambda r: (
            r.setResponseCode(404),
            r.setHeader("content-type", "image/bmp"),
            r.write(b"Image content type error"),
            r.finish(),
        ),
    )

    response = page.request.get(server.PREFIX + "/text-content-type")
    with pytest.raises(AssertionError) as excinfo:
        expect(response).to_be_ok()
    error_message = str(excinfo.value)
    assert ("→ GET " + server.PREFIX + "/text-content-type") in error_message
    assert "← 404 Not Found" in error_message
    assert "Response Text:" in error_message
    assert "Text error" in error_message

    response = page.request.get(server.PREFIX + "/no-content-type")
    with pytest.raises(AssertionError) as excinfo:
        expect(response).to_be_ok()
    error_message = str(excinfo.value)
    assert ("→ GET " + server.PREFIX + "/no-content-type") in error_message
    assert "← 404 Not Found" in error_message
    assert "Response Text:" not in error_message
    assert "No content type error" not in error_message

    response = page.request.get(server.PREFIX + "/binary-content-type")
    with pytest.raises(AssertionError) as excinfo:
        expect(response).to_be_ok()
    error_message = str(excinfo.value)
    assert ("→ GET " + server.PREFIX + "/binary-content-type") in error_message
    assert "← 404 Not Found" in error_message
    assert "Response Text:" not in error_message
    assert "Image content type error" not in error_message


def test_should_print_users_message_for_page_based_assertion(
    page: Page, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<title>new title</title>")
    with pytest.raises(AssertionError) as excinfo:
        expect(page, "Title is not new").to_have_title("old title", timeout=100)
    assert "Title is not new" in str(excinfo.value)
    with pytest.raises(AssertionError) as excinfo:
        expect(page).to_have_title("old title", timeout=100)
    assert "Page title expected to be" in str(excinfo.value)


def test_should_print_expected_value_with_custom_message(
    page: Page, server: Server
) -> None:
    page.goto(server.EMPTY_PAGE)
    page.set_content("<title>new title</title>")
    with pytest.raises(AssertionError) as excinfo:
        expect(page, "custom-message").to_have_title("old title", timeout=100)
    assert "custom-message" in str(excinfo.value)
    assert "Expected value: 'old title'" in str(excinfo.value)
    with pytest.raises(AssertionError) as excinfo:
        expect(page.get_by_text("hello"), "custom-message").to_be_visible(timeout=100)
    assert "custom-message" in str(excinfo.value)
    assert "Expected value" not in str(excinfo.value)


def test_should_be_attached_default(page: Page) -> None:
    page.set_content("<input></input>")
    locator = page.locator("input")
    expect(locator).to_be_attached()


def test_should_be_attached_with_hidden_element(page: Page) -> None:
    page.set_content('<button style="display:none">hello</button>')
    locator = page.locator("button")
    expect(locator).to_be_attached()


def test_should_be_attached_with_not(page: Page) -> None:
    page.set_content("<button>hello</button>")
    locator = page.locator("input")
    expect(locator).not_to_be_attached()


def test_should_be_attached_with_attached_true(page: Page) -> None:
    page.set_content("<button>hello</button>")
    locator = page.locator("button")
    expect(locator).to_be_attached(attached=True)


def test_should_be_attached_with_attached_false(page: Page) -> None:
    page.set_content("<button>hello</button>")
    locator = page.locator("input")
    expect(locator).to_be_attached(attached=False)


def test_should_be_attached_with_attached_false_and_throw_good_error(
    page: Page,
) -> None:
    page.set_content("<button>hello</button>")
    locator = page.locator("button")
    with pytest.raises(AssertionError, match="Locator expected to be detached"):
        expect(locator).to_be_attached(attached=False, timeout=1)


def test_should_be_attached_with_not_and_attached_false(page: Page) -> None:
    page.set_content("<button>hello</button>")
    locator = page.locator("button")
    expect(locator).not_to_be_attached(attached=False)


def test_should_be_attached_eventually(page: Page) -> None:
    page.set_content("<div></div>")
    locator = page.locator("span")
    page.locator("div").evaluate(
        "(e) => setTimeout(() => e.innerHTML = '<span>hello</span>', 1000)"
    )
    expect(locator).to_be_attached()


def test_should_be_attached_eventually_with_not(page: Page) -> None:
    page.set_content("<div><span>Hello</span></div>")
    locator = page.locator("span")
    page.locator("div").evaluate("(e) => setTimeout(() => e.textContent = '', 1000)")
    expect(locator).not_to_be_attached()


def test_should_be_attached_fail(page: Page) -> None:
    page.set_content("<button>Hello</button>")
    locator = page.locator("input")
    with pytest.raises(
        AssertionError, match="Locator expected to be attached"
    ) as exc_info:
        expect(locator).to_be_attached(timeout=1000)
    assert "locator resolved to" not in exc_info.value.args[0]


def test_should_be_attached_fail_with_not(page: Page) -> None:
    page.set_content("<input></input>")
    locator = page.locator("input")
    with pytest.raises(AssertionError) as exc_info:
        expect(locator).not_to_be_attached(timeout=1000)
    assert "locator resolved to <input/>" in exc_info.value.args[0]


def test_should_be_attached_with_impossible_timeout(page: Page) -> None:
    page.set_content("<div id=node>Text content</div>")
    expect(page.locator("#node")).to_be_attached(timeout=1)


def test_should_be_attached_with_impossible_timeout_not(page: Page) -> None:
    page.set_content("<div id=node>Text content</div>")
    expect(page.locator("no-such-thing")).not_to_be_attached(timeout=1)


def test_should_be_able_to_set_custom_timeout(page: Page) -> None:
    with pytest.raises(AssertionError) as exc_info:
        expect(page.locator("#a1")).to_be_visible(timeout=111)
    assert 'Expect "to_be_visible" with timeout 111ms' in str(exc_info.value)


def test_should_be_able_to_set_custom_global_timeout(page: Page) -> None:
    try:
        expect.set_options(timeout=111)
        with pytest.raises(AssertionError) as exc_info:
            expect(page.locator("#a1")).to_be_visible()
        assert 'Expect "to_be_visible" with timeout 111ms' in str(exc_info.value)
    finally:
        expect.set_options(timeout=5_000)


def test_to_have_accessible_name(page: Page) -> None:
    page.set_content('<div role="button" aria-label="Hello"></div>')
    locator = page.locator("div")
    expect(locator).to_have_accessible_name("Hello")
    expect(locator).not_to_have_accessible_name("hello")
    expect(locator).to_have_accessible_name("hello", ignore_case=True)
    expect(locator).to_have_accessible_name(re.compile(r"ell\w"))
    expect(locator).not_to_have_accessible_name(re.compile(r"hello"))
    expect(locator).to_have_accessible_name(re.compile(r"hello"), ignore_case=True)

    page.set_content("<button>foo&nbsp;bar\nbaz</button>")
    expect(page.locator("button")).to_have_accessible_name("foo bar baz")


def test_to_have_accessible_error_message(page: Page) -> None:
    page.set_content(
        """
      <form>
        <input role="textbox" aria-invalid="true" aria-errormessage="error-message" />
        <div id="error-message">Hello</div>
        <div id="irrelevant-error">This should not be considered.</div>
      </form>
    """
    )

    locator = page.locator('input[role="textbox"]')
    expect(locator).to_have_accessible_error_message("Hello")
    expect(locator).not_to_have_accessible_error_message("hello")
    expect(locator).to_have_accessible_error_message("hello", ignore_case=True)
    expect(locator).to_have_accessible_error_message(re.compile(r"ell\w"))
    expect(locator).not_to_have_accessible_error_message(re.compile(r"hello"))
    expect(locator).to_have_accessible_error_message(
        re.compile(r"hello"), ignore_case=True
    )
    expect(locator).not_to_have_accessible_error_message(
        "This should not be considered."
    )


def test_to_have_accessible_error_message_should_handle_multiple_aria_error_message_references(
    page: Page,
) -> None:
    page.set_content(
        """
      <form>
        <input role="textbox" aria-invalid="true" aria-errormessage="error1 error2" />
        <div id="error1">First error message.</div>
        <div id="error2">Second error message.</div>
        <div id="irrelevant-error">This should not be considered.</div>
      </form>
    """
    )

    locator = page.locator('input[role="textbox"]')

    expect(locator).to_have_accessible_error_message(
        "First error message. Second error message."
    )
    expect(locator).to_have_accessible_error_message(
        re.compile(r"first error message.", re.IGNORECASE)
    )
    expect(locator).to_have_accessible_error_message(
        re.compile(r"second error message.", re.IGNORECASE)
    )
    expect(locator).not_to_have_accessible_error_message(
        re.compile(r"This should not be considered.", re.IGNORECASE)
    )


def test_to_have_accessible_description(page: Page) -> None:
    page.set_content('<div role="button" aria-description="Hello"></div>')
    locator = page.locator("div")
    expect(locator).to_have_accessible_description("Hello")
    expect(locator).not_to_have_accessible_description("hello")
    expect(locator).to_have_accessible_description("hello", ignore_case=True)
    expect(locator).to_have_accessible_description(re.compile(r"ell\w"))
    expect(locator).not_to_have_accessible_description(re.compile(r"hello"))
    expect(locator).to_have_accessible_description(
        re.compile(r"hello"), ignore_case=True
    )

    page.set_content(
        """
        <div role="button" aria-describedby="desc"></div>
        <span id="desc">foo&nbsp;bar\nbaz</span>
    """
    )
    expect(page.locator("div")).to_have_accessible_description("foo bar baz")


def test_to_have_role(page: Page) -> None:
    page.set_content('<div role="button">Button!</div>')
    expect(page.locator("div")).to_have_role("button")
    expect(page.locator("div")).not_to_have_role("checkbox")
    with pytest.raises(Error) as excinfo:
        expect(page.locator("div")).to_have_role(re.compile(r"button|checkbox"))  # type: ignore
    assert '"role" argument in to_have_role must be a string' in str(excinfo.value)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
import re
import zipfile
from pathlib import Path
from typing import Any, cast

import pytest

from playwright.sync_api import Browser, BrowserContext, Error, Page, Route, expect
from tests.server import Server


def test_should_work(browser: Browser, server: Server, tmp_path: Path) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(record_har_path=path)
    page = context.new_page()
    page.goto(server.EMPTY_PAGE)
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data


def test_should_omit_content(browser: Browser, server: Server, tmp_path: Path) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(record_har_path=path, record_har_content="omit")
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]

        content1 = log["entries"][0]["response"]["content"]
        assert "text" not in content1
        assert "encoding" not in content1


def test_should_omit_content_legacy(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(record_har_path=path, record_har_omit_content=True)
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]

        content1 = log["entries"][0]["response"]["content"]
        assert "text" not in content1
        assert "encoding" not in content1


def test_should_attach_content(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    path = os.path.join(tmp_path, "log.har.zip")
    context = browser.new_context(
        record_har_path=path,
        record_har_content="attach",
    )
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    page.evaluate("() => fetch('/pptr.png').then(r => r.arrayBuffer())")
    context.close()
    with zipfile.ZipFile(path) as z:
        with z.open("har.har") as har:
            entries = json.load(har)["log"]["entries"]

            assert "encoding" not in entries[0]["response"]["content"]
            assert (
                entries[0]["response"]["content"]["mimeType"]
                == "text/html; charset=utf-8"
            )
            assert (
                "75841480e2606c03389077304342fac2c58ccb1b"
                in entries[0]["response"]["content"]["_file"]
            )
            assert entries[0]["response"]["content"]["size"] >= 96
            assert entries[0]["response"]["content"]["compression"] == 0

            assert "encoding" not in entries[1]["response"]["content"]
            assert (
                entries[1]["response"]["content"]["mimeType"]
                == "text/css; charset=utf-8"
            )
            assert (
                "79f739d7bc88e80f55b9891a22bf13a2b4e18adb"
                in entries[1]["response"]["content"]["_file"]
            )
            assert entries[1]["response"]["content"]["size"] >= 37
            assert entries[1]["response"]["content"]["compression"] == 0

            assert "encoding" not in entries[2]["response"]["content"]
            assert entries[2]["response"]["content"]["mimeType"] == "image/png"
            assert (
                "a4c3a18f0bb83f5d9fe7ce561e065c36205762fa"
                in entries[2]["response"]["content"]["_file"]
            )
            assert entries[2]["response"]["content"]["size"] >= 6000
            assert entries[2]["response"]["content"]["compression"] == 0

            with z.open("75841480e2606c03389077304342fac2c58ccb1b.html") as f:
                assert b"HAR Page" in f.read()

            with z.open("79f739d7bc88e80f55b9891a22bf13a2b4e18adb.css") as f:
                assert b"pink" in f.read()

            with z.open("a4c3a18f0bb83f5d9fe7ce561e065c36205762fa.png") as f:
                assert len(f.read()) == entries[2]["response"]["content"]["size"]


def test_should_include_content(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(record_har_path=path)
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]

        content1 = log["entries"][0]["response"]["content"]
        assert content1["mimeType"] == "text/html; charset=utf-8"
        assert "HAR Page" in content1["text"]


def test_should_default_to_full_mode(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(
        record_har_path=path,
    )
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]
        assert log["entries"][0]["request"]["bodySize"] >= 0


def test_should_support_minimal_mode(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(
        record_har_path=path,
        record_har_mode="minimal",
    )
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]
        assert log["entries"][0]["request"]["bodySize"] == -1


def test_should_filter_by_glob(browser: Browser, server: Server, tmp_path: str) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(
        base_url=server.PREFIX,
        record_har_path=path,
        record_har_url_filter="/*.css",
        ignore_https_errors=True,
    )
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]
        assert len(log["entries"]) == 1
        assert log["entries"][0]["request"]["url"].endswith("one-style.css")


def test_should_filter_by_regexp(
    browser: Browser, server: Server, tmp_path: str
) -> None:
    path = os.path.join(tmp_path, "log.har")
    context = browser.new_context(
        base_url=server.PREFIX,
        record_har_path=path,
        record_har_url_filter=re.compile("HAR.X?HTML", re.I),
        ignore_https_errors=True,
    )
    page = context.new_page()
    page.goto(server.PREFIX + "/har.html")
    context.close()
    with open(path) as f:
        data = json.load(f)
        assert "log" in data
        log = data["log"]
        assert len(log["entries"]) == 1
        assert log["entries"][0]["request"]["url"].endswith("har.html")


def test_should_context_route_from_har_matching_the_method_and_following_redirects(
    context: BrowserContext, assetdir: Path
) -> None:
    context.route_from_har(har=assetdir / "har-fulfill.har")
    page = context.new_page()
    page.goto("http://no.playwright/")
    # HAR contains a redirect for the script that should be followed automatically.
    assert page.evaluate("window.value") == "foo"
    # HAR contains a POST for the css file that should not be used.
    expect(page.locator("body")).to_have_css("background-color", "rgb(255, 0, 0)")


def test_should_page_route_from_har_matching_the_method_and_following_redirects(
    page: Page, assetdir: Path
) -> None:
    page.route_from_har(har=assetdir / "har-fulfill.har")
    page.goto("http://no.playwright/")
    # HAR contains a redirect for the script that should be followed automatically.
    assert page.evaluate("window.value") == "foo"
    # HAR contains a POST for the css file that should not be used.
    expect(page.locator("body")).to_have_css("background-color", "rgb(255, 0, 0)")


def test_fallback_continue_should_continue_when_not_found_in_har(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(har=assetdir / "har-fulfill.har", not_found="fallback")
    page = context.new_page()
    page.goto(server.PREFIX + "/one-style.html")
    expect(page.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")


def test_by_default_should_abort_requests_not_found_in_har(
    context: BrowserContext,
    server: Server,
    assetdir: Path,
    is_chromium: bool,
    is_webkit: bool,
) -> None:
    context.route_from_har(har=assetdir / "har-fulfill.har")
    page = context.new_page()

    with pytest.raises(Error) as exc_info:
        page.goto(server.EMPTY_PAGE)
    assert exc_info.value
    if is_chromium:
        assert "net::ERR_FAILED" in exc_info.value.message
    elif is_webkit:
        assert "Blocked by Web Inspector" in exc_info.value.message
    else:
        assert "NS_ERROR_FAILURE" in exc_info.value.message


def test_fallback_continue_should_continue_requests_on_bad_har(
    context: BrowserContext, server: Server, tmp_path: Path
) -> None:
    path_to_invalid_har = tmp_path / "invalid.har"
    with path_to_invalid_har.open("w") as f:
        json.dump({"log": {}}, f)
    context.route_from_har(har=path_to_invalid_har, not_found="fallback")
    page = context.new_page()
    page.goto(server.PREFIX + "/one-style.html")
    expect(page.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")


def test_should_only_handle_requests_matching_url_filter(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(
        har=assetdir / "har-fulfill.har", not_found="fallback", url="**/*.js"
    )
    page = context.new_page()

    def handler(route: Route) -> None:
        assert route.request.url == "http://no.playwright/"
        route.fulfill(
            status=200,
            content_type="text/html",
            body='<script src="./script.js"></script><div>hello</div>',
        )

    context.route("http://no.playwright/", handler)
    page.goto("http://no.playwright/")
    assert page.evaluate("window.value") == "foo"
    expect(page.locator("body")).to_have_css("background-color", "rgba(0, 0, 0, 0)")


def test_should_only_handle_requests_matching_url_filter_no_fallback(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(har=assetdir / "har-fulfill.har", url="**/*.js")
    page = context.new_page()

    def handler(route: Route) -> None:
        assert route.request.url == "http://no.playwright/"
        route.fulfill(
            status=200,
            content_type="text/html",
            body='<script src="./script.js"></script><div>hello</div>',
        )

    context.route("http://no.playwright/", handler)
    page.goto("http://no.playwright/")
    assert page.evaluate("window.value") == "foo"
    expect(page.locator("body")).to_have_css("background-color", "rgba(0, 0, 0, 0)")


def test_should_only_handle_requests_matching_url_filter_no_fallback_page(
    page: Page, server: Server, assetdir: Path
) -> None:
    page.route_from_har(har=assetdir / "har-fulfill.har", url="**/*.js")

    def handler(route: Route) -> None:
        assert route.request.url == "http://no.playwright/"
        route.fulfill(
            status=200,
            content_type="text/html",
            body='<script src="./script.js"></script><div>hello</div>',
        )

    page.route("http://no.playwright/", handler)
    page.goto("http://no.playwright/")
    assert page.evaluate("window.value") == "foo"
    expect(page.locator("body")).to_have_css("background-color", "rgba(0, 0, 0, 0)")


def test_should_support_regex_filter(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(
        har=assetdir / "har-fulfill.har",
        url=re.compile(r".*(\.js|.*\.css|no.playwright\/)"),
    )
    page = context.new_page()
    page.goto("http://no.playwright/")
    assert page.evaluate("window.value") == "foo"
    expect(page.locator("body")).to_have_css("background-color", "rgb(255, 0, 0)")


def test_should_go_back_to_redirected_navigation(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(
        har=assetdir / "har-redirect.har", url=re.compile(r"/.*theverge.*/")
    )
    page = context.new_page()
    page.goto("https://theverge.com/")
    page.goto(server.EMPTY_PAGE)
    expect(page).to_have_url(server.EMPTY_PAGE)

    response = page.go_back()
    assert response
    expect(page).to_have_url("https://www.theverge.com/")
    assert response.request.url == "https://www.theverge.com/"
    assert page.evaluate("window.location.href") == "https://www.theverge.com/"


@pytest.mark.skip_browser(
    "firefox"
)  # skipped upstream (https://github.com/microsoft/playwright/blob/6a8d835145e2f4002ee00b67a80a1f70af956703/tests/library/browsercontext-har.spec.ts#L214)
def test_should_go_forward_to_redirected_navigation(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(
        har=assetdir / "har-redirect.har", url=re.compile(r"/.*theverge.*/")
    )
    page = context.new_page()
    page.goto("https://theverge.com/")
    page.goto(server.EMPTY_PAGE)
    expect(page).to_have_url(server.EMPTY_PAGE)
    page.goto("https://theverge.com/")
    expect(page).to_have_url("https://www.theverge.com/")
    page.go_back()
    expect(page).to_have_url(server.EMPTY_PAGE)
    response = page.go_forward()
    assert response
    expect(page).to_have_url("https://www.theverge.com/")
    assert response.request.url == "https://www.theverge.com/"
    assert page.evaluate("window.location.href") == "https://www.theverge.com/"


def test_should_reload_redirected_navigation(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(
        har=assetdir / "har-redirect.har", url=re.compile(r"/.*theverge.*/")
    )
    page = context.new_page()
    page.goto("https://theverge.com/")
    expect(page).to_have_url("https://www.theverge.com/")
    response = page.reload()
    assert response
    expect(page).to_have_url("https://www.theverge.com/")
    assert response.request.url == "https://www.theverge.com/"
    assert page.evaluate("window.location.href") == "https://www.theverge.com/"


def test_should_fulfill_from_har_with_content_in_a_file(
    context: BrowserContext, server: Server, assetdir: Path
) -> None:
    context.route_from_har(har=assetdir / "har-sha1.har")
    page = context.new_page()
    page.goto("http://no.playwright/")
    assert page.content() == "<html><head></head><body>Hello, world</body></html>"


def test_should_round_trip_har_zip(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    har_path = tmp_path / "har.zip"
    context_1 = browser.new_context(record_har_mode="minimal", record_har_path=har_path)
    page_1 = context_1.new_page()
    page_1.goto(server.PREFIX + "/one-style.html")
    context_1.close()

    context_2 = browser.new_context()
    context_2.route_from_har(har=har_path, not_found="abort")
    page_2 = context_2.new_page()
    page_2.goto(server.PREFIX + "/one-style.html")
    assert "hello, world!" in page_2.content()
    expect(page_2.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")
    context_2.close()


def test_should_round_trip_har_with_post_data(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    server.set_route(
        "/echo", lambda req: (req.write(cast(Any, req).post_body), req.finish())
    )
    fetch_function = """
        async (body) => {
            const response = await fetch('/echo', { method: 'POST', body });
            return response.text();
        };
    """
    har_path = tmp_path / "har.zip"
    context_1 = browser.new_context(record_har_mode="minimal", record_har_path=har_path)
    page_1 = context_1.new_page()
    page_1.goto(server.EMPTY_PAGE)

    assert page_1.evaluate(fetch_function, "1") == "1"
    assert page_1.evaluate(fetch_function, "2") == "2"
    assert page_1.evaluate(fetch_function, "3") == "3"
    context_1.close()

    context_2 = browser.new_context()
    context_2.route_from_har(har=har_path, not_found="abort")
    page_2 = context_2.new_page()
    page_2.goto(server.EMPTY_PAGE)
    assert page_2.evaluate(fetch_function, "1") == "1"
    assert page_2.evaluate(fetch_function, "2") == "2"
    assert page_2.evaluate(fetch_function, "3") == "3"
    with pytest.raises(Exception):
        page_2.evaluate(fetch_function, "4")
    context_2.close()


def test_should_disambiguate_by_header(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    server.set_route(
        "/echo",
        lambda req: (req.write(cast(str, req.getHeader("baz")).encode()), req.finish()),
    )
    fetch_function = """
        async (bazValue) => {
            const response = await fetch('/echo', {
            method: 'POST',
            body: '',
            headers: {
                foo: 'foo-value',
                bar: 'bar-value',
                baz: bazValue,
            }
            });
            return response.text();
        };
    """
    har_path = tmp_path / "har.zip"
    context_1 = browser.new_context(record_har_mode="minimal", record_har_path=har_path)
    page_1 = context_1.new_page()
    page_1.goto(server.EMPTY_PAGE)

    assert page_1.evaluate(fetch_function, "baz1") == "baz1"
    assert page_1.evaluate(fetch_function, "baz2") == "baz2"
    assert page_1.evaluate(fetch_function, "baz3") == "baz3"
    context_1.close()

    context_2 = browser.new_context()
    context_2.route_from_har(har=har_path)
    page_2 = context_2.new_page()
    page_2.goto(server.EMPTY_PAGE)
    assert page_2.evaluate(fetch_function, "baz1") == "baz1"
    assert page_2.evaluate(fetch_function, "baz2") == "baz2"
    assert page_2.evaluate(fetch_function, "baz3") == "baz3"
    assert page_2.evaluate(fetch_function, "baz4") == "baz1"
    context_2.close()


def test_should_produce_extracted_zip(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    har_path = tmp_path / "har.har"
    context = browser.new_context(
        record_har_mode="minimal", record_har_path=har_path, record_har_content="attach"
    )
    page_1 = context.new_page()
    page_1.goto(server.PREFIX + "/one-style.html")
    context.close()

    assert har_path.exists()
    with har_path.open() as r:
        content = r.read()
        assert "log" in content
        assert "background-color" not in r.read()

    context_2 = browser.new_context()
    context_2.route_from_har(har_path, not_found="abort")
    page_2 = context_2.new_page()
    page_2.goto(server.PREFIX + "/one-style.html")
    assert "hello, world!" in page_2.content()
    expect(page_2.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")
    context_2.close()


def test_should_update_har_zip_for_context(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    har_path = tmp_path / "har.zip"
    context = browser.new_context()
    context.route_from_har(har_path, update=True)
    page_1 = context.new_page()
    page_1.goto(server.PREFIX + "/one-style.html")
    context.close()

    assert har_path.exists()

    context_2 = browser.new_context()
    context_2.route_from_har(har_path, not_found="abort")
    page_2 = context_2.new_page()
    page_2.goto(server.PREFIX + "/one-style.html")
    assert "hello, world!" in page_2.content()
    expect(page_2.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")
    context_2.close()


def test_should_update_har_zip_for_page(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    har_path = tmp_path / "har.zip"
    context = browser.new_context()
    page_1 = context.new_page()
    page_1.route_from_har(har_path, update=True)
    page_1.goto(server.PREFIX + "/one-style.html")
    context.close()

    assert har_path.exists()

    context_2 = browser.new_context()
    page_2 = context_2.new_page()
    page_2.route_from_har(har_path, not_found="abort")
    page_2.goto(server.PREFIX + "/one-style.html")
    assert "hello, world!" in page_2.content()
    expect(page_2.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")
    context_2.close()


def test_should_update_har_zip_for_page_with_different_options(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    har_path = tmp_path / "har.zip"
    context1 = browser.new_context()
    page1 = context1.new_page()
    page1.route_from_har(
        har_path, update=True, update_content="embed", update_mode="full"
    )
    page1.goto(server.PREFIX + "/one-style.html")
    context1.close()

    context2 = browser.new_context()
    page2 = context2.new_page()
    page2.route_from_har(har_path, not_found="abort")
    page2.goto(server.PREFIX + "/one-style.html")
    assert "hello, world!" in page2.content()
    expect(page2.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")
    context2.close()


def test_should_update_extracted_har_zip_for_page(
    browser: Browser, server: Server, tmp_path: Path
) -> None:
    har_path = tmp_path / "har.har"
    context = browser.new_context()
    page_1 = context.new_page()
    page_1.route_from_har(har_path, update=True)
    page_1.goto(server.PREFIX + "/one-style.html")
    context.close()

    assert har_path.exists()
    with har_path.open() as r:
        content = r.read()
        assert "log" in content
        assert "background-color" not in r.read()

    context_2 = browser.new_context()
    page_2 = context_2.new_page()
    page_2.route_from_har(har_path, not_found="abort")
    page_2.goto(server.PREFIX + "/one-style.html")
    assert "hello, world!" in page_2.content()
    expect(page_2.locator("body")).to_have_css("background-color", "rgb(255, 192, 203)")
    context_2.close()



#!/usr/bin/env python
# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from pathlib import Path
from typing import List

LICENSE_HEADER = """
# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""".strip()


def file_has_license(file_path: str) -> None:
    path = Path(file_path)
    if not path.name.endswith(".py"):
        return
    if not path.exists():
        raise Exception(f"File {file_path} does not exist")
    content = path.read_text(encoding="utf-8")
    if content.strip() == "":
        return
    if content.startswith("#!"):
        # flake8: noqa: E203
        content = content[content.find("\n") + 1 :]
    if not content.startswith(LICENSE_HEADER):
        raise Exception(f"File {file_path} does not have license header")


def main() -> None:
    errors: List[str] = []

    for file_path in sys.argv[1:]:
        try:
            file_has_license(file_path)
        except Exception as e:
            errors.append(str(e))

    if errors:
        print("Missing license header in files:")
        for error in errors:
            print("  -", error)
        sys.exit(1)


if __name__ == "__main__":
    main()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import time
from pathlib import Path
from typing import Any, Callable

import pytest

from playwright.sync_api import BrowserType, Error, Playwright, Route
from tests.conftest import RemoteServer
from tests.server import Server


def test_browser_type_connect_slow_mo(
    server: Server, browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    browser = browser_type.connect(remote_server.ws_endpoint, slow_mo=100)
    browser_context = browser.new_context()
    t1 = time.monotonic()
    page = browser_context.new_page()
    assert page.evaluate("11 * 11") == 121
    assert (time.monotonic() - t1) >= 0.2
    page.goto(server.EMPTY_PAGE)
    browser.close()


def test_browser_type_connect_should_be_able_to_reconnect_to_a_browser(
    server: Server, browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    browser = browser_type.connect(remote_server.ws_endpoint)
    browser_context = browser.new_context()
    assert len(browser_context.pages) == 0
    page = browser_context.new_page()
    assert len(browser_context.pages) == 1
    assert page.evaluate("11 * 11") == 121
    page.goto(server.EMPTY_PAGE)
    browser.close()

    browser = browser_type.connect(remote_server.ws_endpoint)
    browser_context = browser.new_context()
    page = browser_context.new_page()
    page.goto(server.EMPTY_PAGE)
    browser.close()


def test_browser_type_connect_should_be_able_to_connect_two_browsers_at_the_same_time(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    browser1 = browser_type.connect(remote_server.ws_endpoint)
    assert len(browser1.contexts) == 0
    browser1.new_context()
    assert len(browser1.contexts) == 1

    browser2 = browser_type.connect(remote_server.ws_endpoint)
    assert len(browser2.contexts) == 0
    browser2.new_context()
    assert len(browser2.contexts) == 1
    assert len(browser1.contexts) == 1

    browser1.close()
    page2 = browser2.new_page()
    # original browser should still work
    assert page2.evaluate("7 * 6") == 42

    browser2.close()


def test_browser_type_connect_disconnected_event_should_be_emitted_when_browser_is_closed_or_server_is_closed(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    # Launch another server to not affect other tests.
    remote = launch_server()

    browser1 = browser_type.connect(remote.ws_endpoint)
    browser2 = browser_type.connect(remote.ws_endpoint)

    disconnected1 = []
    disconnected2 = []
    browser1.on("disconnected", lambda browser: disconnected1.append(True))
    browser2.on("disconnected", lambda browser: disconnected2.append(True))

    page2 = browser2.new_page()

    browser1.close()
    assert len(disconnected1) == 1
    assert len(disconnected2) == 0

    remote.kill()
    assert len(disconnected1) == 1

    with pytest.raises(Error):
        # Tickle connection so that it gets a chance to dispatch disconnect event.
        page2.title()
    assert len(disconnected2) == 1


def test_browser_type_disconnected_event_should_have_browser_as_argument(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    browser = browser_type.connect(remote_server.ws_endpoint)
    event_payloads = []
    browser.on("disconnected", lambda b: event_payloads.append(b))
    browser.close()
    assert event_payloads[0] == browser


def test_browser_type_connect_set_browser_connected_state(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    browser = browser_type.connect(remote_server.ws_endpoint)
    assert browser.is_connected()
    browser.close()
    assert browser.is_connected() is False


def test_browser_type_connect_should_throw_when_used_after_is_connected_returns_false(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    browser = browser_type.connect(remote_server.ws_endpoint)
    page = browser.new_page()

    remote_server.kill()

    with pytest.raises(Error) as exc_info:
        page.evaluate("1 + 1")
    assert "has been closed" in exc_info.value.message
    assert browser.is_connected() is False


def test_browser_type_connect_should_forward_close_events_to_pages(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    # Launch another server to not affect other tests.
    remote = launch_server()

    browser = browser_type.connect(remote.ws_endpoint)
    context = browser.new_context()
    page = context.new_page()

    events = []
    browser.on("disconnected", lambda browser: events.append("browser::disconnected"))
    context.on("close", lambda context: events.append("context::close"))
    page.on("close", lambda page: events.append("page::close"))

    browser.close()
    assert events == ["page::close", "context::close", "browser::disconnected"]
    remote.kill()
    assert events == ["page::close", "context::close", "browser::disconnected"]


def test_browser_type_connect_should_forward_close_events_on_remote_kill(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    # Launch another server to not affect other tests.
    remote = launch_server()

    browser = browser_type.connect(remote.ws_endpoint)
    context = browser.new_context()
    page = context.new_page()

    events = []
    browser.on("disconnected", lambda context: events.append("browser::disconnected"))
    context.on("close", lambda context: events.append("context::close"))
    page.on("close", lambda page: events.append("page::close"))
    remote.kill()
    with pytest.raises(Error):
        page.title()
    assert events == ["page::close", "context::close", "browser::disconnected"]


def test_connect_to_closed_server_without_hangs(
    browser_type: BrowserType, launch_server: Callable[[], RemoteServer]
) -> None:
    remote_server = launch_server()
    remote_server.kill()
    with pytest.raises(Error) as exc:
        browser_type.connect(remote_server.ws_endpoint)
    assert "WebSocket error: " in exc.value.message


def test_browser_type_connect_should_fulfill_with_global_fetch_result(
    browser_type: BrowserType,
    launch_server: Callable[[], RemoteServer],
    playwright: Playwright,
    server: Server,
) -> None:
    # Launch another server to not affect other tests.
    remote = launch_server()

    browser = browser_type.connect(remote.ws_endpoint)
    context = browser.new_context()
    page = context.new_page()

    def handle_request(route: Route) -> None:
        request = playwright.request.new_context()
        response = request.get(server.PREFIX + "/simple.json")
        return route.fulfill(response=response)

    page.route("**/*", handle_request)

    response = page.goto(server.EMPTY_PAGE)
    assert response
    assert response.status == 200
    assert response.json() == {"foo": "bar"}

    remote.kill()


def test_set_input_files_should_preserve_last_modified_timestamp(
    browser_type: BrowserType,
    launch_server: Callable[[], RemoteServer],
    assetdir: Path,
) -> None:
    # Launch another server to not affect other tests.
    remote = launch_server()

    browser = browser_type.connect(remote.ws_endpoint)
    context = browser.new_context()
    page = context.new_page()

    page.set_content("<input type=file multiple=true/>")
    input = page.locator("input")
    files: Any = ["file-to-upload.txt", "file-to-upload-2.txt"]
    input.set_input_files([assetdir / file for file in files])
    assert input.evaluate("input => [...input.files].map(f => f.name)") == files
    timestamps = input.evaluate("input => [...input.files].map(f => f.lastModified)")
    expected_timestamps = [os.path.getmtime(assetdir / file) * 1000 for file in files]

    # On Linux browser sometimes reduces the timestamp by 1ms: 1696272058110.0715  -> 1696272058109 or even
    # rounds it to seconds in WebKit: 1696272058110 -> 1696272058000.
    for i in range(len(timestamps)):
        assert abs(timestamps[i] - expected_timestamps[i]) < 1000



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from playwright.async_api import Error, JSHandle, Page
from tests.server import Server

from .utils import Utils


async def captureLastKeydown(page: Page) -> JSHandle:
    lastEvent = await page.evaluate_handle(
        """() => {
        const lastEvent = {
        repeat: false,
        location: -1,
        code: '',
        key: '',
        metaKey: false,
        keyIdentifier: 'unsupported'
        };
        document.addEventListener('keydown', e => {
        lastEvent.repeat = e.repeat;
        lastEvent.location = e.location;
        lastEvent.key = e.key;
        lastEvent.code = e.code;
        lastEvent.metaKey = e.metaKey;
        // keyIdentifier only exists in WebKit, and isn't in TypeScript's lib.
        lastEvent.keyIdentifier = 'keyIdentifier' in e && e.keyIdentifier;
        }, true);
        return lastEvent;
    }"""
    )
    return lastEvent


async def test_keyboard_type_into_a_textarea(page: Page) -> None:
    await page.evaluate(
        """
            const textarea = document.createElement('textarea');
            document.body.appendChild(textarea);
            textarea.focus();
        """
    )
    text = "Hello world. I am the text that was typed!"
    await page.keyboard.type(text)
    assert await page.evaluate('document.querySelector("textarea").value') == text


async def test_keyboard_move_with_the_arrow_keys(page: Page, server: Server) -> None:
    await page.goto(f"{server.PREFIX}/input/textarea.html")
    await page.type("textarea", "Hello World!")
    assert (
        await page.evaluate("document.querySelector('textarea').value")
        == "Hello World!"
    )
    for _ in "World!":
        await page.keyboard.press("ArrowLeft")
    await page.keyboard.type("inserted ")
    assert (
        await page.evaluate("document.querySelector('textarea').value")
        == "Hello inserted World!"
    )
    await page.keyboard.down("Shift")
    for _ in "inserted ":
        await page.keyboard.press("ArrowLeft")
    await page.keyboard.up("Shift")
    await page.keyboard.press("Backspace")
    assert (
        await page.evaluate("document.querySelector('textarea').value")
        == "Hello World!"
    )


async def test_keyboard_send_a_character_with_elementhandle_press(
    page: Page, server: Server
) -> None:
    await page.goto(f"{server.PREFIX}/input/textarea.html")
    textarea = await page.query_selector("textarea")
    assert textarea
    await textarea.press("a")
    assert await page.evaluate("document.querySelector('textarea').value") == "a"
    await page.evaluate(
        "() => window.addEventListener('keydown', e => e.preventDefault(), true)"
    )
    await textarea.press("b")
    assert await page.evaluate("document.querySelector('textarea').value") == "a"


async def test_should_send_a_character_with_send_character(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    await page.focus("textarea")
    await page.keyboard.insert_text("嗨")
    assert await page.evaluate('() => document.querySelector("textarea").value') == "嗨"
    await page.evaluate(
        '() => window.addEventListener("keydown", e => e.preventDefault(), true)'
    )
    await page.keyboard.insert_text("a")
    assert (
        await page.evaluate('() => document.querySelector("textarea").value') == "嗨a"
    )


async def test_should_only_emit_input_event(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    await page.focus("textarea")
    events = await page.evaluate_handle(
        """() => {
    const events = [];
    document.addEventListener('keydown', e => events.push(e.type));
    document.addEventListener('keyup', e => events.push(e.type));
    document.addEventListener('keypress', e => events.push(e.type));
    document.addEventListener('input', e => events.push(e.type));
    return events;
  }"""
    )

    await page.keyboard.insert_text("hello world")
    assert await events.json_value() == ["input"]


async def test_should_report_shiftkey(
    page: Page, server: Server, is_mac: bool, is_firefox: bool
) -> None:
    if is_firefox and is_mac:
        pytest.skip()
    await page.goto(server.PREFIX + "/input/keyboard.html")
    keyboard = page.keyboard
    codeForKey = {"Shift": 16, "Alt": 18, "Control": 17}
    for modifierKey in codeForKey.keys():
        await keyboard.down(modifierKey)
        assert (
            await page.evaluate("() => getResult()")
            == "Keydown: "
            + modifierKey
            + " "
            + modifierKey
            + "Left "
            + str(codeForKey[modifierKey])
            + " ["
            + modifierKey
            + "]"
        )
        await keyboard.down("!")
        # Shift+! will generate a keypress
        if modifierKey == "Shift":
            assert (
                await page.evaluate("() => getResult()")
                == "Keydown: ! Digit1 49 ["
                + modifierKey
                + "]\nKeypress: ! Digit1 33 33 ["
                + modifierKey
                + "]"
            )
        else:
            assert (
                await page.evaluate("() => getResult()")
                == "Keydown: ! Digit1 49 [" + modifierKey + "]"
            )

        await keyboard.up("!")
        assert (
            await page.evaluate("() => getResult()")
            == "Keyup: ! Digit1 49 [" + modifierKey + "]"
        )
        await keyboard.up(modifierKey)
        assert (
            await page.evaluate("() => getResult()")
            == "Keyup: "
            + modifierKey
            + " "
            + modifierKey
            + "Left "
            + str(codeForKey[modifierKey])
            + " []"
        )


async def test_should_report_multiple_modifiers(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    keyboard = page.keyboard
    await keyboard.down("Control")
    assert (
        await page.evaluate("() => getResult()")
        == "Keydown: Control ControlLeft 17 [Control]"
    )
    await keyboard.down("Alt")
    assert (
        await page.evaluate("() => getResult()")
        == "Keydown: Alt AltLeft 18 [Alt Control]"
    )
    await keyboard.down(";")
    assert (
        await page.evaluate("() => getResult()")
        == "Keydown: ; Semicolon 186 [Alt Control]"
    )
    await keyboard.up(";")
    assert (
        await page.evaluate("() => getResult()")
        == "Keyup: ; Semicolon 186 [Alt Control]"
    )
    await keyboard.up("Control")
    assert (
        await page.evaluate("() => getResult()")
        == "Keyup: Control ControlLeft 17 [Alt]"
    )
    await keyboard.up("Alt")
    assert await page.evaluate("() => getResult()") == "Keyup: Alt AltLeft 18 []"


async def test_should_send_proper_codes_while_typing(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    await page.keyboard.type("!")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: ! Digit1 49 []",
            "Keypress: ! Digit1 33 33 []",
            "Keyup: ! Digit1 49 []",
        ]
    )
    await page.keyboard.type("^")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: ^ Digit6 54 []",
            "Keypress: ^ Digit6 94 94 []",
            "Keyup: ^ Digit6 54 []",
        ]
    )


async def test_should_send_proper_codes_while_typing_with_shift(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    keyboard = page.keyboard
    await keyboard.down("Shift")
    await page.keyboard.type("~")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: Shift ShiftLeft 16 [Shift]",
            "Keydown: ~ Backquote 192 [Shift]",  # 192 is ` keyCode
            "Keypress: ~ Backquote 126 126 [Shift]",  # 126 is ~ charCode
            "Keyup: ~ Backquote 192 [Shift]",
        ]
    )
    await keyboard.up("Shift")


async def test_should_not_type_canceled_events(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    await page.focus("textarea")
    await page.evaluate(
        """() => {
    window.addEventListener('keydown', event => {
      event.stopPropagation();
      event.stopImmediatePropagation();
      if (event.key === 'l')
        event.preventDefault();
      if (event.key === 'o')
        event.preventDefault();
    }, false);
  }"""
    )

    await page.keyboard.type("Hello World!")
    assert (
        await page.eval_on_selector("textarea", "textarea => textarea.value")
        == "He Wrd!"
    )


async def test_should_press_plus(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    await page.keyboard.press("+")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: + Equal 187 []",  # 192 is ` keyCode
            "Keypress: + Equal 43 43 []",  # 126 is ~ charCode
            "Keyup: + Equal 187 []",
        ]
    )


async def test_should_press_shift_plus(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    await page.keyboard.press("Shift++")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: Shift ShiftLeft 16 [Shift]",
            "Keydown: + Equal 187 [Shift]",  # 192 is ` keyCode
            "Keypress: + Equal 43 43 [Shift]",  # 126 is ~ charCode
            "Keyup: + Equal 187 [Shift]",
            "Keyup: Shift ShiftLeft 16 []",
        ]
    )


async def test_should_support_plus_separated_modifiers(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    await page.keyboard.press("Shift+~")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: Shift ShiftLeft 16 [Shift]",
            "Keydown: ~ Backquote 192 [Shift]",  # 192 is ` keyCode
            "Keypress: ~ Backquote 126 126 [Shift]",  # 126 is ~ charCode
            "Keyup: ~ Backquote 192 [Shift]",
            "Keyup: Shift ShiftLeft 16 []",
        ]
    )


async def test_should_suport_multiple_plus_separated_modifiers(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    await page.keyboard.press("Control+Shift+~")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: Control ControlLeft 17 [Control]",
            "Keydown: Shift ShiftLeft 16 [Control Shift]",
            "Keydown: ~ Backquote 192 [Control Shift]",  # 192 is ` keyCode
            "Keyup: ~ Backquote 192 [Control Shift]",
            "Keyup: Shift ShiftLeft 16 [Control]",
            "Keyup: Control ControlLeft 17 []",
        ]
    )


async def test_should_shift_raw_codes(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/keyboard.html")
    await page.keyboard.press("Shift+Digit3")
    assert await page.evaluate("() => getResult()") == "\n".join(
        [
            "Keydown: Shift ShiftLeft 16 [Shift]",
            "Keydown: # Digit3 51 [Shift]",  # 51 is # keyCode
            "Keypress: # Digit3 35 35 [Shift]",  # 35 is # charCode
            "Keyup: # Digit3 51 [Shift]",
            "Keyup: Shift ShiftLeft 16 []",
        ]
    )


async def test_should_specify_repeat_property(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    await page.focus("textarea")
    lastEvent = await captureLastKeydown(page)
    await page.keyboard.down("a")
    assert await lastEvent.evaluate("e => e.repeat") is False
    await page.keyboard.press("a")
    assert await lastEvent.evaluate("e => e.repeat")

    await page.keyboard.down("b")
    assert await lastEvent.evaluate("e => e.repeat") is False
    await page.keyboard.down("b")
    assert await lastEvent.evaluate("e => e.repeat")

    await page.keyboard.up("a")
    await page.keyboard.down("a")
    assert await lastEvent.evaluate("e => e.repeat") is False


async def test_should_type_all_kinds_of_characters(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    await page.focus("textarea")
    text = "This text goes onto two lines.\nThis character is 嗨."
    await page.keyboard.type(text)
    assert await page.eval_on_selector("textarea", "t => t.value") == text


async def test_should_specify_location(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    lastEvent = await captureLastKeydown(page)
    textarea = await page.query_selector("textarea")
    assert textarea

    await textarea.press("Digit5")
    assert await lastEvent.evaluate("e => e.location") == 0

    await textarea.press("ControlLeft")
    assert await lastEvent.evaluate("e => e.location") == 1

    await textarea.press("ControlRight")
    assert await lastEvent.evaluate("e => e.location") == 2

    await textarea.press("NumpadSubtract")
    assert await lastEvent.evaluate("e => e.location") == 3


async def test_should_press_enter(page: Page) -> None:
    await page.set_content("<textarea></textarea>")
    await page.focus("textarea")
    lastEventHandle = await captureLastKeydown(page)

    async def testEnterKey(key: str, expectedKey: str, expectedCode: str) -> None:
        await page.keyboard.press(key)
        lastEvent = await lastEventHandle.json_value()
        assert lastEvent["key"] == expectedKey
        assert lastEvent["code"] == expectedCode
        value = await page.eval_on_selector("textarea", "t => t.value")
        assert value == "\n"
        await page.eval_on_selector("textarea", "t => t.value = ''")

    await testEnterKey("Enter", "Enter", "Enter")
    await testEnterKey("NumpadEnter", "Enter", "NumpadEnter")
    await testEnterKey("\n", "Enter", "Enter")
    await testEnterKey("\r", "Enter", "Enter")


async def test_should_throw_unknown_keys(page: Page, server: Server) -> None:
    with pytest.raises(Error) as exc:
        await page.keyboard.press("NotARealKey")
    assert exc.value.message == 'Keyboard.press: Unknown key: "NotARealKey"'

    with pytest.raises(Error) as exc:
        await page.keyboard.press("ё")
    assert exc.value.message == 'Keyboard.press: Unknown key: "ё"'

    with pytest.raises(Error) as exc:
        await page.keyboard.press("😊")
    assert exc.value.message == 'Keyboard.press: Unknown key: "😊"'


async def test_should_type_emoji(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    await page.type("textarea", "👹 Tokyo street Japan 🇯🇵")
    assert (
        await page.eval_on_selector("textarea", "textarea => textarea.value")
        == "👹 Tokyo street Japan 🇯🇵"
    )


async def test_should_type_emoji_into_an_iframe(
    page: Page, server: Server, utils: Utils
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await utils.attach_frame(page, "emoji-test", server.PREFIX + "/input/textarea.html")
    frame = page.frames[1]
    textarea = await frame.query_selector("textarea")
    assert textarea
    await textarea.type("👹 Tokyo street Japan 🇯🇵")
    assert (
        await frame.eval_on_selector("textarea", "textarea => textarea.value")
        == "👹 Tokyo street Japan 🇯🇵"
    )


async def test_should_handle_select_all(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    textarea = await page.query_selector("textarea")
    assert textarea
    await textarea.type("some text")
    await page.keyboard.down("ControlOrMeta")
    await page.keyboard.press("a")
    await page.keyboard.up("ControlOrMeta")
    await page.keyboard.press("Backspace")
    assert await page.eval_on_selector("textarea", "textarea => textarea.value") == ""


async def test_should_be_able_to_prevent_select_all(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    textarea = await page.query_selector("textarea")
    assert textarea
    await textarea.type("some text")
    await page.eval_on_selector(
        "textarea",
        """textarea => {
    textarea.addEventListener('keydown', event => {
      if (event.key === 'a' && (event.metaKey || event.ctrlKey))
        event.preventDefault();
    }, false);
  }""",
    )

    await page.keyboard.down("ControlOrMeta")
    await page.keyboard.press("a")
    await page.keyboard.up("ControlOrMeta")
    await page.keyboard.press("Backspace")
    assert (
        await page.eval_on_selector("textarea", "textarea => textarea.value")
        == "some tex"
    )


@pytest.mark.only_platform("darwin")
@pytest.mark.skip_browser("firefox")  # Upstream issue
async def test_should_support_macos_shortcuts(
    page: Page, server: Server, is_firefox: bool, is_mac: bool
) -> None:
    await page.goto(server.PREFIX + "/input/textarea.html")
    textarea = await page.query_selector("textarea")
    assert textarea
    await textarea.type("some text")
    # select one word backwards
    await page.keyboard.press("Shift+Control+Alt+KeyB")
    await page.keyboard.press("Backspace")
    assert (
        await page.eval_on_selector("textarea", "textarea => textarea.value") == "some "
    )


async def test_should_press_the_meta_key(page: Page) -> None:
    lastEvent = await captureLastKeydown(page)
    await page.keyboard.press("Meta")
    v = await lastEvent.json_value()
    metaKey = v["metaKey"]
    key = v["key"]
    code = v["code"]
    assert key == "Meta"
    assert code == "MetaLeft"
    assert metaKey


async def test_should_work_after_a_cross_origin_navigation(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/empty.html")
    await page.goto(server.CROSS_PROCESS_PREFIX + "/empty.html")
    lastEvent = await captureLastKeydown(page)
    await page.keyboard.press("a")
    assert await lastEvent.evaluate("l => l.key") == "a"


# event.keyIdentifier has been removed from all browsers except WebKit
@pytest.mark.only_browser("webkit")
async def test_should_expose_keyIdentifier_in_webkit(
    page: Page, server: Server
) -> None:
    lastEvent = await captureLastKeydown(page)
    keyMap = {
        "ArrowUp": "Up",
        "ArrowDown": "Down",
        "ArrowLeft": "Left",
        "ArrowRight": "Right",
        "Backspace": "U+0008",
        "Tab": "U+0009",
        "Delete": "U+007F",
        "a": "U+0041",
        "b": "U+0042",
        "F12": "F12",
    }
    for key, keyIdentifier in keyMap.items():
        await page.keyboard.press(key)
        assert await lastEvent.evaluate("e => e.keyIdentifier") == keyIdentifier


async def test_should_scroll_with_pagedown(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/input/scrollable.html")
    # A click is required for WebKit to send the event into the body.
    await page.click("body")
    await page.keyboard.press("PageDown")
    # We can't wait for the scroll to finish, so just wait for it to start.
    await page.wait_for_function("() => scrollY > 0")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import pytest

from playwright.sync_api import BrowserContext, Dialog, Page

from ..server import Server, TestServerRequest


def test_console_event_should_work(page: Page) -> None:
    with page.context.expect_console_message() as console_info:
        page.evaluate("() => console.log('hello')")
    message = console_info.value
    assert message.text == "hello"
    assert message.page == page


def test_console_event_should_work_in_popup(page: Page) -> None:
    with page.context.expect_console_message() as console_info:
        with page.expect_popup() as popup_info:
            page.evaluate(
                """() => {
                const win = window.open('');
                win.console.log('hello');
                }"""
            )
    message = console_info.value
    popup = popup_info.value
    assert message.text == "hello"
    assert message.page == popup


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
def test_console_event_should_work_in_popup_2(page: Page, browser_name: str) -> None:
    with page.context.expect_console_message(
        lambda msg: msg.type == "log"
    ) as console_info:
        with page.context.expect_page() as page_info:
            page.evaluate(
                """async () => {
                const win = window.open('javascript:console.log("hello")');
                await new Promise(f => setTimeout(f, 0));
                win.close();
            }"""
            )
    message = console_info.value
    popup = page_info.value
    assert message.text == "hello"
    assert message.page == popup


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
def test_console_event_should_work_in_immediately_closed_popup(
    page: Page, browser_name: str
) -> None:
    with page.context.expect_console_message(
        lambda msg: msg.type == "log"
    ) as console_info:
        with page.context.expect_page() as page_info:
            page.evaluate(
                """() => {
                const win = window.open('');
                win.console.log('hello');
                win.close();
            }"""
            )
    message = console_info.value
    popup = page_info.value
    assert message.text == "hello"
    assert message.page == popup


def test_dialog_event_should_work1(page: Page) -> None:
    dialog1: Optional[Dialog] = None

    def handle_page_dialog(dialog: Dialog) -> None:
        nonlocal dialog1
        dialog1 = dialog
        dialog.accept("hello")

    page.on("dialog", handle_page_dialog)

    dialog2: Optional[Dialog] = None

    def handle_context_dialog(dialog: Dialog) -> None:
        nonlocal dialog2
        dialog2 = dialog

    page.context.on("dialog", handle_context_dialog)

    assert page.evaluate("() => prompt('hey?')") == "hello"
    assert dialog1
    assert dialog1 == dialog2
    assert dialog1.message == "hey?"
    assert dialog1.page == page


def test_dialog_event_should_work_in_popup1(page: Page) -> None:
    dialog: Optional[Dialog] = None

    def handle_dialog(d: Dialog) -> None:
        nonlocal dialog
        dialog = d
        dialog.accept("hello")

    page.context.on("dialog", handle_dialog)

    with page.expect_popup() as popup_info:
        assert page.evaluate("() => window.open('').prompt('hey?')") == "hello"
    popup = popup_info.value
    assert dialog
    assert dialog.message == "hey?"
    assert dialog.page == popup


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
def test_dialog_event_should_work_in_popup_2(page: Page, browser_name: str) -> None:
    def handle_dialog(dialog: Dialog) -> None:
        assert dialog.message == "hey?"
        assert dialog.page is None
        dialog.accept("hello")

    page.context.on("dialog", handle_dialog)

    assert page.evaluate("() => window.open('javascript:prompt(\"hey?\")')")


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
def test_dialog_event_should_work_in_immdiately_closed_popup(page: Page) -> None:
    popup = None

    def handle_popup(p: Page) -> None:
        nonlocal popup
        popup = p

    page.on("popup", handle_popup)

    with page.context.expect_console_message() as console_info:
        page.evaluate(
            """() => {
                const win = window.open();
                win.console.log('hello');
                win.close();
            }"""
        )
    message = console_info.value

    assert message.text == "hello"
    assert message.page == popup


def test_dialog_event_should_work_with_inline_script_tag(
    page: Page, server: Server
) -> None:
    def handle_route(request: TestServerRequest) -> None:
        request.setHeader("content-type", "text/html")
        request.write(b"<script>window.result = prompt('hey?')</script>")
        request.finish()

    server.set_route("/popup.html", handle_route)
    page.goto(server.EMPTY_PAGE)
    page.set_content("<a href='popup.html' target=_blank>Click me</a>")

    with (
        page.context.expect_event("dialog") as dialog_info,
        page.expect_popup() as popup_info,
    ):
        page.click("a")

    dialog: Dialog = dialog_info.value
    popup: Page = popup_info.value

    assert dialog.message == "hey?"
    assert dialog.page == popup
    dialog.accept("hello")
    assert popup.evaluate("window.result") == "hello"


def test_console_event_should_work_with_context_manager(page: Page) -> None:
    with page.context.expect_console_message() as cm_info:
        page.evaluate("() => console.log('hello')")
    message = cm_info.value
    assert message.text == "hello"
    assert message.page == page


def test_weberror_event_should_work(context: BrowserContext, page: Page) -> None:
    with context.expect_event("weberror") as error_info:
        page.goto('data:text/html,<script>throw new Error("Test")</script>')
    error = error_info.value
    assert error.page == page
    assert error.error.message == "Test"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re

import pytest

from playwright.async_api import Error, Page, expect


async def test_has_text_and_internal_text_should_match_full_node_text_in_strict_mode(
    page: Page,
) -> None:
    await page.set_content(
        """
        <div id=div1>hello<span>world</span></div>
        <div id=div2>hello</div>
    """
    )
    await expect(page.get_by_text("helloworld", exact=True)).to_have_id("div1")
    await expect(page.get_by_text("hello", exact=True)).to_have_id("div2")
    await expect(page.locator("div", has_text=re.compile("^helloworld$"))).to_have_id(
        "div1"
    )
    await expect(page.locator("div", has_text=re.compile("^hello$"))).to_have_id("div2")

    await page.set_content(
        """
        <div id=div1><span id=span1>hello</span>world</div>
        <div id=div2><span id=span2>hello</span></div>
    """
    )
    await expect(page.get_by_text("helloworld", exact=True)).to_have_id("div1")
    assert await page.get_by_text("hello", exact=True).evaluate_all(
        "els => els.map(e => e.id)"
    ) == ["span1", "span2"]
    await expect(page.locator("div", has_text=re.compile("^helloworld$"))).to_have_id(
        "div1"
    )
    await expect(page.locator("div", has_text=re.compile("^hello$"))).to_have_id("div2")


async def test_should_work(page: Page) -> None:
    await page.set_content(
        """
        <div>yo</div><div>ya</div><div>\nye  </div>
    """
    )
    assert await page.eval_on_selector("text=ya", "e => e.outerHTML") == "<div>ya</div>"
    assert (
        await page.eval_on_selector('text="ya"', "e => e.outerHTML") == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("text=/^[ay]+$/", "e => e.outerHTML")
        == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("text=/Ya/i", "e => e.outerHTML") == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("text=ye", "e => e.outerHTML")
        == "<div>\nye  </div>"
    )
    assert ">\nye  </div>" in await page.get_by_text("ye").evaluate("e => e.outerHTML")

    await page.set_content(
        """
        <div> ye </div><div>ye</div>
    """
    )
    assert (
        await page.eval_on_selector('text="ye"', "e => e.outerHTML")
        == "<div> ye </div>"
    )
    assert "> ye </div>" in await page.get_by_text("ye", exact=True).first.evaluate(
        "e => e.outerHTML"
    )

    await page.set_content(
        """
        <div>yo</div><div>"ya</div><div> hello world! </div>
    """
    )
    assert (
        await page.eval_on_selector('text="\\"ya"', "e => e.outerHTML")
        == '<div>"ya</div>'
    )
    assert (
        await page.eval_on_selector("text=/hello/", "e => e.outerHTML")
        == "<div> hello world! </div>"
    )
    assert (
        await page.eval_on_selector("text=/^\\s*heLLo/i", "e => e.outerHTML")
        == "<div> hello world! </div>"
    )

    await page.set_content(
        """
        <div>yo<div>ya</div>hey<div>hey</div></div>
    """
    )
    assert (
        await page.eval_on_selector("text=hey", "e => e.outerHTML") == "<div>hey</div>"
    )
    assert (
        await page.eval_on_selector('text=yo>>text="ya"', "e => e.outerHTML")
        == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector('text=yo>> text="ya"', "e => e.outerHTML")
        == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("text=yo >>text='ya'", "e => e.outerHTML")
        == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("text=yo >> text='ya'", "e => e.outerHTML")
        == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("'yo'>>\"ya\"", "e => e.outerHTML")
        == "<div>ya</div>"
    )
    assert (
        await page.eval_on_selector("\"yo\" >> 'ya'", "e => e.outerHTML")
        == "<div>ya</div>"
    )

    await page.set_content(
        """
        <div>yo<span id="s1"></span></div><div>yo<span id="s2"></span><span id="s3"></span></div>
        """
    )
    assert (
        await page.eval_on_selector_all(
            "text=yo", "es => es.map(e => e.outerHTML).join('\\n')"
        )
        == '<div>yo<span id="s1"></span></div>\n<div>yo<span id="s2"></span><span id="s3"></span></div>'
    )

    await page.set_content("<div>'</div><div>\"</div><div>\\</div><div>x</div>")
    assert (
        await page.eval_on_selector("text='\\''", "e => e.outerHTML") == "<div>'</div>"
    )
    assert (
        await page.eval_on_selector("text='\"'", "e => e.outerHTML") == '<div>"</div>'
    )
    assert (
        await page.eval_on_selector('text="\\""', "e => e.outerHTML") == '<div>"</div>'
    )
    assert (
        await page.eval_on_selector('text="\'"', "e => e.outerHTML") == "<div>'</div>"
    )
    assert (
        await page.eval_on_selector('text="\\x"', "e => e.outerHTML") == "<div>x</div>"
    )
    assert (
        await page.eval_on_selector("text='\\x'", "e => e.outerHTML") == "<div>x</div>"
    )
    assert (
        await page.eval_on_selector("text='\\\\'", "e => e.outerHTML")
        == "<div>\\</div>"
    )
    assert (
        await page.eval_on_selector('text="\\\\"', "e => e.outerHTML")
        == "<div>\\</div>"
    )
    assert await page.eval_on_selector('text="', "e => e.outerHTML") == '<div>"</div>'
    assert await page.eval_on_selector("text='", "e => e.outerHTML") == "<div>'</div>"
    assert await page.eval_on_selector('"x"', "e => e.outerHTML") == "<div>x</div>"
    assert await page.eval_on_selector("'x'", "e => e.outerHTML") == "<div>x</div>"
    with pytest.raises(Error):
        await page.query_selector_all('"')
    with pytest.raises(Error):
        await page.query_selector_all("'")

    await page.set_content("<div> ' </div><div> \" </div>")
    assert await page.eval_on_selector('text="', "e => e.outerHTML") == '<div> " </div>'
    assert await page.eval_on_selector("text='", "e => e.outerHTML") == "<div> ' </div>"

    await page.set_content("<div>Hi''&gt;&gt;foo=bar</div>")
    assert (
        await page.eval_on_selector("text=\"Hi''>>foo=bar\"", "e => e.outerHTML")
        == "<div>Hi''&gt;&gt;foo=bar</div>"
    )
    await page.set_content("<div>Hi'\"&gt;&gt;foo=bar</div>")
    assert (
        await page.eval_on_selector('text="Hi\'\\">>foo=bar"', "e => e.outerHTML")
        == "<div>Hi'\"&gt;&gt;foo=bar</div>"
    )

    await page.set_content("<div>Hi&gt;&gt;<span></span></div>")
    assert (
        await page.eval_on_selector('text="Hi>>">>span', "e => e.outerHTML")
        == "<span></span>"
    )
    assert (
        await page.eval_on_selector("text=/Hi\\>\\>/ >> span", "e => e.outerHTML")
        == "<span></span>"
    )

    await page.set_content("<div>a<br>b</div><div>a</div>")
    assert (
        await page.eval_on_selector("text=a", "e => e.outerHTML") == "<div>a<br>b</div>"
    )
    assert (
        await page.eval_on_selector("text=b", "e => e.outerHTML") == "<div>a<br>b</div>"
    )
    assert (
        await page.eval_on_selector("text=ab", "e => e.outerHTML")
        == "<div>a<br>b</div>"
    )
    assert await page.query_selector("text=abc") is None
    assert await page.eval_on_selector_all("text=a", "els => els.length") == 2
    assert await page.eval_on_selector_all("text=b", "els => els.length") == 1
    assert await page.eval_on_selector_all("text=ab", "els => els.length") == 1
    assert await page.eval_on_selector_all("text=abc", "els => els.length") == 0

    await page.set_content("<div></div><span></span>")
    await page.eval_on_selector(
        "div",
        """div => {
        div.appendChild(document.createTextNode('hello'))
        div.appendChild(document.createTextNode('world'))
    }""",
    )
    await page.eval_on_selector(
        "span",
        """span => {
        span.appendChild(document.createTextNode('hello'))
        span.appendChild(document.createTextNode('world'))
    }""",
    )
    assert (
        await page.eval_on_selector("text=lowo", "e => e.outerHTML")
        == "<div>helloworld</div>"
    )
    assert (
        await page.eval_on_selector_all(
            "text=lowo", "els => els.map(e => e.outerHTML).join('')"
        )
        == "<div>helloworld</div><span>helloworld</span>"
    )

    await page.set_content("<span>Sign&nbsp;in</span><span>Hello\n \nworld</span>")
    assert (
        await page.eval_on_selector("text=Sign in", "e => e.outerHTML")
        == "<span>Sign&nbsp;in</span>"
    )
    assert len((await page.query_selector_all("text=Sign \tin"))) == 1
    assert len((await page.query_selector_all('text="Sign in"'))) == 1
    assert (
        await page.eval_on_selector("text=lo wo", "e => e.outerHTML")
        == "<span>Hello\n \nworld</span>"
    )
    assert (
        await page.eval_on_selector('text="Hello world"', "e => e.outerHTML")
        == "<span>Hello\n \nworld</span>"
    )
    assert await page.query_selector('text="lo wo"') is None
    assert len((await page.query_selector_all("text=lo \nwo"))) == 1
    assert len(await page.query_selector_all('text="lo \nwo"')) == 0

    await page.set_content("<div>let's<span>hello</span></div>")
    assert (
        await page.eval_on_selector("text=/let's/i >> span", "e => e.outerHTML")
        == "<span>hello</span>"
    )
    assert (
        await page.eval_on_selector("text=/let\\'s/i >> span", "e => e.outerHTML")
        == "<span>hello</span>"
    )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from playwright.sync_api import Page

TODO_ITEMS = ["buy some cheese", "feed the cat", "book a doctors appointment"]


def create_default_todos(page: Page) -> None:
    for item in TODO_ITEMS:
        page.locator(".new-todo").fill(item)
        page.locator(".new-todo").press("Enter")


def check_number_of_completed_todos_in_local_storage(page: Page, expected: int) -> None:
    assert (
        page.evaluate(
            "JSON.parse(localStorage['react-todos']).filter(i => i.completed).length"
        )
        == expected
    )


def assert_number_of_todos_in_local_storage(page: Page, expected: int) -> None:
    assert len(page.evaluate("JSON.parse(localStorage['react-todos'])")) == expected


def check_todos_in_local_storage(page: Page, title: str) -> None:
    assert title in page.evaluate(
        "JSON.parse(localStorage['react-todos']).map(i => i.title)"
    )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import multiprocessing
import os
import signal
import sys
from typing import Any, Dict

import pytest

from playwright.async_api import async_playwright
from playwright.sync_api import sync_playwright


def _test_signals_async(
    browser_name: str, launch_arguments: Dict, wait_queue: "multiprocessing.Queue[str]"
) -> None:
    # On Windows, hint to mypy and pyright that they shouldn't check this function
    if sys.platform == "win32":
        return

    os.setpgrp()
    sigint_received = False

    def my_sig_handler(signum: int, frame: Any) -> None:
        nonlocal sigint_received
        sigint_received = True

    signal.signal(signal.SIGINT, my_sig_handler)

    async def main() -> None:
        playwright = await async_playwright().start()
        browser = await playwright[browser_name].launch(
            **launch_arguments,
            handle_sigint=False,
        )
        context = await browser.new_context()
        page = await context.new_page()
        notified = False
        try:
            while not sigint_received:
                if not notified:
                    wait_queue.put("ready")
                    notified = True
                await page.wait_for_timeout(100)
        finally:
            wait_queue.put("close context")
            await context.close()
            wait_queue.put("close browser")
            await browser.close()
            wait_queue.put("close playwright")
            await playwright.stop()
            wait_queue.put("all done")

    asyncio.run(main())


def _test_signals_sync(
    browser_name: str, launch_arguments: Dict, wait_queue: "multiprocessing.Queue[str]"
) -> None:
    # On Windows, hint to mypy and pyright that they shouldn't check this function
    if sys.platform == "win32":
        return

    os.setpgrp()
    sigint_received = False

    def my_sig_handler(signum: int, frame: Any) -> None:
        nonlocal sigint_received
        sigint_received = True

    signal.signal(signal.SIGINT, my_sig_handler)

    playwright = sync_playwright().start()
    browser = playwright[browser_name].launch(
        **launch_arguments,
        handle_sigint=False,
    )
    context = browser.new_context()
    page = context.new_page()
    notified = False
    try:
        while not sigint_received:
            if not notified:
                wait_queue.put("ready")
                notified = True
            page.wait_for_timeout(100)
    finally:
        wait_queue.put("close context")
        context.close()
        wait_queue.put("close browser")
        browser.close()
        wait_queue.put("close playwright")
        playwright.stop()
        wait_queue.put("all done")


def _create_signals_test(
    target: Any, browser_name: str, launch_arguments: Dict
) -> None:
    # On Windows, hint to mypy and pyright that they shouldn't check this function
    if sys.platform == "win32":
        return

    wait_queue: "multiprocessing.Queue[str]" = multiprocessing.Queue()
    process = multiprocessing.Process(
        target=target, args=[browser_name, launch_arguments, wait_queue]
    )
    process.start()
    assert process.pid is not None
    logs = [wait_queue.get()]
    os.killpg(os.getpgid(process.pid), signal.SIGINT)
    process.join()
    while not wait_queue.empty():
        logs.append(wait_queue.get())
    assert logs == [
        "ready",
        "close context",
        "close browser",
        "close playwright",
        "all done",
    ]
    assert process.exitcode == 0


@pytest.mark.skipif(sys.platform == "win32", reason="there is no SIGINT on Windows")
def test_signals_sync(browser_name: str, launch_arguments: Dict) -> None:
    _create_signals_test(_test_signals_sync, browser_name, launch_arguments)


@pytest.mark.skipif(sys.platform == "win32", reason="there is no SIGINT on Windows")
def test_signals_async(browser_name: str, launch_arguments: Dict) -> None:
    _create_signals_test(_test_signals_async, browser_name, launch_arguments)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from playwright.sync_api import BrowserContext, Page
from tests.server import Server
from tests.utils import must


def test_context_unroute_all_removes_all_handlers(
    page: Page, context: BrowserContext, server: Server
) -> None:
    context.route(
        "**/*",
        lambda route: route.abort(),
    )
    context.route(
        "**/empty.html",
        lambda route: route.abort(),
    )
    context.unroute_all()
    page.goto(server.EMPTY_PAGE)


def test_page_unroute_all_removes_all_routes(page: Page, server: Server) -> None:
    page.route(
        "**/*",
        lambda route: route.abort(),
    )
    page.route(
        "**/empty.html",
        lambda route: route.abort(),
    )
    page.unroute_all()
    response = must(page.goto(server.EMPTY_PAGE))
    assert response.ok



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Dict

import pytest

from playwright.sync_api import BrowserType, Error


@pytest.mark.skip_browser("firefox")
def test_browser_type_launch_should_throw_if_page_argument_is_passed(
    browser_type: BrowserType, launch_arguments: Dict
) -> None:
    with pytest.raises(Error) as exc:
        browser_type.launch(**launch_arguments, args=["http://example.com"])
    assert "can not specify page" in exc.value.message


def test_browser_type_launch_should_reject_if_launched_browser_fails_immediately(
    browser_type: BrowserType, launch_arguments: Dict, assetdir: Path
) -> None:
    with pytest.raises(Error):
        browser_type.launch(
            **launch_arguments,
            executable_path=assetdir / "dummy_bad_browser_executable.js",
        )


def test_browser_type_launch_should_reject_if_executable_path_is_invalid(
    browser_type: BrowserType, launch_arguments: Dict
) -> None:
    with pytest.raises(Error) as exc:
        browser_type.launch(**launch_arguments, executable_path="random-invalid-path")
    assert "executable doesn't exist" in exc.value.message


def test_browser_type_executable_path_should_work(
    browser_type: BrowserType, browser_channel: str
) -> None:
    if browser_channel:
        return
    executable_path = browser_type.executable_path
    assert os.path.exists(executable_path)
    assert os.path.realpath(executable_path) == os.path.realpath(executable_path)


def test_browser_type_name_should_work(
    browser_type: BrowserType, is_webkit: bool, is_firefox: bool, is_chromium: bool
) -> None:
    if is_webkit:
        assert browser_type.name == "webkit"
    elif is_firefox:
        assert browser_type.name == "firefox"
    elif is_chromium:
        assert browser_type.name == "chromium"
    else:
        raise ValueError("Unknown browser")


def test_browser_close_should_fire_close_event_for_all_contexts(
    browser_type: BrowserType, launch_arguments: Dict
) -> None:
    browser = browser_type.launch(**launch_arguments)
    context = browser.new_context()
    closed = []
    context.on("close", lambda _: closed.append(True))
    browser.close()
    assert closed == [True]


def test_browser_close_should_be_callable_twice(
    browser_type: BrowserType, launch_arguments: Dict
) -> None:
    browser = browser_type.launch(**launch_arguments)
    browser.close()
    browser.close()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import AsyncGenerator, Optional, cast

import pytest

from playwright.async_api import Browser, BrowserContext, ElementHandle, JSHandle, Page


@pytest.fixture
async def context(browser: Browser) -> AsyncGenerator[BrowserContext, None]:
    context = await browser.new_context(has_touch=True)
    yield context
    await context.close()


async def test_should_send_all_of_the_correct_events(page: Page) -> None:
    await page.set_content(
        """
            <div id="a" style="background: lightblue; width: 50px; height: 50px">a</div>
            <div id="b" style="background: pink; width: 50px; height: 50px">b</div>
        """
    )
    await page.tap("#a")
    element_handle = await track_events(await page.query_selector("#b"))
    await page.tap("#b")
    assert await element_handle.json_value() == [
        "pointerover",
        "pointerenter",
        "pointerdown",
        "touchstart",
        "pointerup",
        "pointerout",
        "pointerleave",
        "touchend",
        "mouseover",
        "mouseenter",
        "mousemove",
        "mousedown",
        "mouseup",
        "click",
    ]


async def test_should_not_send_mouse_events_touchstart_is_canceled(page: Page) -> None:
    await page.set_content("hello world")
    await page.evaluate(
        """() => {
            // touchstart is not cancelable unless passive is false
            document.addEventListener('touchstart', t => t.preventDefault(), {passive: false});
        }"""
    )
    events_handle = await track_events(await page.query_selector("body"))
    await page.tap("body")
    assert await events_handle.json_value() == [
        "pointerover",
        "pointerenter",
        "pointerdown",
        "touchstart",
        "pointerup",
        "pointerout",
        "pointerleave",
        "touchend",
    ]


async def test_should_not_send_mouse_events_touchend_is_canceled(page: Page) -> None:
    await page.set_content("hello world")
    await page.evaluate(
        """() => {
            // touchstart is not cancelable unless passive is false
            document.addEventListener('touchend', t => t.preventDefault());
        }"""
    )
    events_handle = await track_events(await page.query_selector("body"))
    await page.tap("body")
    assert await events_handle.json_value() == [
        "pointerover",
        "pointerenter",
        "pointerdown",
        "touchstart",
        "pointerup",
        "pointerout",
        "pointerleave",
        "touchend",
    ]


async def test_should_work_with_modifiers(page: Page) -> None:
    await page.set_content("hello world")
    alt_key_promise = asyncio.create_task(
        page.evaluate(
            """() => new Promise(resolve => {
                document.addEventListener('touchstart', event => {
                    resolve(event.altKey);
                }, {passive: false});
            })"""
        )
    )
    await asyncio.sleep(0)  # make sure the evals hit the page
    await page.evaluate("""() => void 0""")
    await page.tap("body", modifiers=["Alt"])
    assert await alt_key_promise is True


async def test_should_send_well_formed_touch_points(page: Page) -> None:
    promises = asyncio.gather(
        page.evaluate(
            """() => new Promise(resolve => {
                    document.addEventListener('touchstart', event => {
                        resolve([...event.touches].map(t => ({
                        identifier: t.identifier,
                        clientX: t.clientX,
                        clientY: t.clientY,
                        pageX: t.pageX,
                        pageY: t.pageY,
                        radiusX: 'radiusX' in t ? t.radiusX : t['webkitRadiusX'],
                        radiusY: 'radiusY' in t ? t.radiusY : t['webkitRadiusY'],
                        rotationAngle: 'rotationAngle' in t ? t.rotationAngle : t['webkitRotationAngle'],
                        force: 'force' in t ? t.force : t['webkitForce'],
                        })));
                    }, false);
                })"""
        ),
        page.evaluate(
            """() => new Promise(resolve => {
                    document.addEventListener('touchend', event => {
                        resolve([...event.touches].map(t => ({
                        identifier: t.identifier,
                        clientX: t.clientX,
                        clientY: t.clientY,
                        pageX: t.pageX,
                        pageY: t.pageY,
                        radiusX: 'radiusX' in t ? t.radiusX : t['webkitRadiusX'],
                        radiusY: 'radiusY' in t ? t.radiusY : t['webkitRadiusY'],
                        rotationAngle: 'rotationAngle' in t ? t.rotationAngle : t['webkitRotationAngle'],
                        force: 'force' in t ? t.force : t['webkitForce'],
                        })));
                    }, false);
                })"""
        ),
    )
    # make sure the evals hit the page
    await page.evaluate("""() => void 0""")
    await page.touchscreen.tap(40, 60)
    [touchstart, touchend] = await promises
    assert touchstart == [
        {
            "clientX": 40,
            "clientY": 60,
            "force": 1,
            "identifier": 0,
            "pageX": 40,
            "pageY": 60,
            "radiusX": 1,
            "radiusY": 1,
            "rotationAngle": 0,
        }
    ]
    assert touchend == []


async def test_should_wait_until_an_element_is_visible_to_tap_it(page: Page) -> None:
    div = cast(
        ElementHandle,
        await page.evaluate_handle(
            """() => {
            const button = document.createElement('button');
            button.textContent = 'not clicked';
            document.body.appendChild(button);
            button.style.display = 'none';
            return button;
        }"""
        ),
    )
    tap_promise = asyncio.create_task(div.tap())
    await asyncio.sleep(0)  # issue tap
    await div.evaluate("""div => div.onclick = () => div.textContent = 'clicked'""")
    await div.evaluate("""div => div.style.display = 'block'""")
    await tap_promise
    assert await div.text_content() == "clicked"


async def test_locators_tap(page: Page) -> None:
    await page.set_content(
        """
        <div id="a" style="background: lightblue; width: 50px; height: 50px">a</div>
        <div id="b" style="background: pink; width: 50px; height: 50px">b</div>
    """
    )
    await page.locator("#a").tap()
    element_handle = await track_events(await page.query_selector("#b"))
    await page.locator("#b").tap()
    assert await element_handle.json_value() == [
        "pointerover",
        "pointerenter",
        "pointerdown",
        "touchstart",
        "pointerup",
        "pointerout",
        "pointerleave",
        "touchend",
        "mouseover",
        "mouseenter",
        "mousemove",
        "mousedown",
        "mouseup",
        "click",
    ]


async def track_events(target: Optional[ElementHandle]) -> JSHandle:
    assert target
    return await target.evaluate_handle(
        """target => {
            const events = [];
            for (const event of [
                'mousedown', 'mouseenter', 'mouseleave', 'mousemove', 'mouseout', 'mouseover', 'mouseup', 'click',
                'pointercancel', 'pointerdown', 'pointerenter', 'pointerleave', 'pointermove', 'pointerout', 'pointerover', 'pointerup',
                'touchstart', 'touchend', 'touchmove', 'touchcancel',])
                target.addEventListener(event, () => events.push(event), false);
            return events;
        }"""
    )



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from playwright.async_api import Page, Route
from tests.server import Server


async def test_should_fetch_original_request_and_fulfill(
    page: Page, server: Server
) -> None:
    async def handle(route: Route) -> None:
        response = await page.request.fetch(route.request)
        await route.fulfill(response=response)

    await page.route("**/*", handle)
    response = await page.goto(server.PREFIX + "/title.html")
    assert response
    assert response.status == 200
    assert await page.title() == "Woof-Woof"


async def test_should_fulfill_json(page: Page, server: Server) -> None:
    async def handle(route: Route) -> None:
        await route.fulfill(status=201, headers={"foo": "bar"}, json={"bar": "baz"})

    await page.route("**/*", handle)

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.status == 201
    assert response.headers["content-type"] == "application/json"
    assert await response.json() == {"bar": "baz"}


async def test_should_fulfill_json_overriding_existing_response(
    page: Page, server: Server
) -> None:
    server.set_route(
        "/tags",
        lambda request: (
            request.setHeader("foo", "bar"),
            request.write('{"tags": ["a", "b"]}'.encode()),
            request.finish(),
        ),
    )

    original = {}

    async def handle(route: Route) -> None:
        response = await route.fetch()
        json = await response.json()
        original["tags"] = json["tags"]
        json["tags"] = ["c"]
        await route.fulfill(response=response, json=json)

    await page.route("**/*", handle)

    response = await page.goto(server.PREFIX + "/tags")
    assert response
    assert response.status == 200
    assert response.headers["content-type"] == "application/json"
    assert response.headers["foo"] == "bar"
    assert original["tags"] == ["a", "b"]
    assert await response.json() == {"tags": ["c"]}



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from typing import Generator

import pytest

from playwright.sync_api import Page, expect

from .utils import (
    TODO_ITEMS,
    assert_number_of_todos_in_local_storage,
    create_default_todos,
)


@pytest.fixture(autouse=True)
def run_around_tests(page: Page) -> Generator[None, None, None]:
    # setup before a test
    page.goto("https://demo.playwright.dev/todomvc")
    # run the actual test
    yield
    # run any cleanup code


def test_new_todo_test_should_allow_me_to_add_todo_items(page: Page) -> None:
    # Create 1st todo.
    page.locator(".new-todo").fill(TODO_ITEMS[0])
    page.locator(".new-todo").press("Enter")

    # Make sure the list only has one todo item.
    expect(page.locator(".view label")).to_have_text([TODO_ITEMS[0]])

    # Create 2nd todo.
    page.locator(".new-todo").fill(TODO_ITEMS[1])
    page.locator(".new-todo").press("Enter")

    # Make sure the list now has two todo items.
    expect(page.locator(".view label")).to_have_text([TODO_ITEMS[0], TODO_ITEMS[1]])

    assert_number_of_todos_in_local_storage(page, 2)


def test_new_todo_test_should_clear_text_input_field_when_an_item_is_added(
    page: Page,
) -> None:
    # Create one todo item.
    page.locator(".new-todo").fill(TODO_ITEMS[0])
    page.locator(".new-todo").press("Enter")

    # Check that input is empty.
    expect(page.locator(".new-todo")).to_be_empty()
    assert_number_of_todos_in_local_storage(page, 1)


def test_new_todo_test_should_append_new_items_to_the_bottom_of_the_list(
    page: Page,
) -> None:
    # Create 3 items.
    create_default_todos(page)

    # Check test using different methods.
    expect(page.locator(".todo-count")).to_have_text("3 items left")
    expect(page.locator(".todo-count")).to_contain_text("3")
    expect(page.locator(".todo-count")).to_have_text(re.compile("3"))

    # Check all items in one call.
    expect(page.locator(".view label")).to_have_text(TODO_ITEMS)
    assert_number_of_todos_in_local_storage(page, 3)


def test_new_todo_should_show_main_and_foter_when_items_added(page: Page) -> None:
    page.locator(".new-todo").fill(TODO_ITEMS[0])
    page.locator(".new-todo").press("Enter")

    expect(page.locator(".main")).to_be_visible()
    expect(page.locator(".footer")).to_be_visible()
    assert_number_of_todos_in_local_storage(page, 1)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import pytest

from playwright.sync_api import Browser, Page
from tests.server import Server


def test_should_work(page: Page, server: Server) -> None:
    with page.expect_event("requestfinished") as request_info:
        page.goto(server.EMPTY_PAGE)
    request = request_info.value
    timing = request.timing
    verify_connections_timing_consistency(timing)
    assert timing["requestStart"] >= timing["connectEnd"]
    assert timing["responseStart"] >= timing["requestStart"]
    assert timing["responseEnd"] >= timing["responseStart"]
    assert timing["responseEnd"] < 10000


def test_should_work_for_subresource(
    page: Page, server: Server, is_win: bool, is_mac: bool, is_webkit: bool
) -> None:
    if is_webkit and (is_mac or is_win):
        pytest.skip()
    requests = []
    page.on("requestfinished", lambda request: requests.append(request))
    page.goto(server.PREFIX + "/one-style.html")
    assert len(requests) >= 2
    timing = requests[1].timing
    verify_connections_timing_consistency(timing)
    assert timing["requestStart"] >= 0
    assert timing["responseStart"] > timing["requestStart"]
    assert timing["responseEnd"] >= timing["responseStart"]
    assert timing["responseEnd"] < 10000


def test_should_work_for_ssl(
    browser: Browser, https_server: Server, is_mac: bool, is_webkit: bool
) -> None:
    if is_webkit and is_mac:
        pytest.skip()
    page = browser.new_page(ignore_https_errors=True)
    with page.expect_event("requestfinished") as request_info:
        page.goto(https_server.EMPTY_PAGE)
    request = request_info.value
    timing = request.timing
    verify_connections_timing_consistency(timing)
    assert timing["requestStart"] >= timing["connectEnd"]
    assert timing["responseStart"] >= timing["requestStart"]
    assert timing["responseEnd"] >= timing["responseStart"]
    assert timing["responseEnd"] < 10000
    page.close()


@pytest.mark.skip_browser("webkit")  # In WebKit, redirects don"t carry the timing info
def test_should_work_for_redirect(page: Page, server: Server) -> None:
    server.set_redirect("/foo.html", "/empty.html")
    responses = []
    page.on("response", lambda response: responses.append(response))
    page.goto(server.PREFIX + "/foo.html")
    for r in responses:
        r.finished()

    assert len(responses) == 2
    assert responses[0].url == server.PREFIX + "/foo.html"
    assert responses[1].url == server.PREFIX + "/empty.html"

    timing1 = responses[0].request.timing
    verify_connections_timing_consistency(timing1)
    assert timing1["requestStart"] >= timing1["connectEnd"]
    assert timing1["responseStart"] > timing1["requestStart"]
    assert timing1["responseEnd"] >= timing1["responseStart"]
    assert timing1["responseEnd"] < 10000

    timing2 = responses[1].request.timing
    verify_connections_timing_consistency(timing2)
    assert timing2["requestStart"] >= 0
    assert timing2["responseStart"] > timing2["requestStart"]
    assert timing2["responseEnd"] >= timing2["responseStart"]
    assert timing2["responseEnd"] < 10000


def verify_timing_value(value: float, previous: float) -> None:
    assert value == -1 or value > 0 and value >= previous


def verify_connections_timing_consistency(timing: Dict) -> None:
    verify_timing_value(timing["domainLookupStart"], -1)
    verify_timing_value(timing["domainLookupEnd"], timing["domainLookupStart"])
    verify_timing_value(timing["connectStart"], timing["domainLookupEnd"])
    verify_timing_value(timing["secureConnectionStart"], timing["connectStart"])
    verify_timing_value(timing["connectEnd"], timing["secureConnectionStart"])



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from playwright.sync_api import Page, expect


def test_get_by_test_id(page: Page) -> None:
    page.set_content("<div><div data-testid='Hello'>Hello world</div></div>")
    expect(page.get_by_test_id("Hello")).to_have_text("Hello world")
    expect(page.main_frame.get_by_test_id("Hello")).to_have_text("Hello world")
    expect(page.locator("div").get_by_test_id("Hello")).to_have_text("Hello world")


def test_get_by_test_id_escape_id(page: Page) -> None:
    page.set_content("<div><div data-testid='He\"llo'>Hello world</div></div>")
    expect(page.get_by_test_id('He"llo')).to_have_text("Hello world")


def test_get_by_text(page: Page) -> None:
    page.set_content("<div><div>yo</div><div>ya</div><div>\nye  </div></div>")

    expect(page.get_by_text("yo")).to_have_count(1)
    expect(page.main_frame.get_by_text("yo")).to_have_count(1)
    expect(page.locator("div").get_by_text("yo")).to_have_count(1)

    assert ">\nye  </div>" in page.get_by_text("ye").evaluate("e => e.outerHTML")
    assert ">\nye  </div>" in page.get_by_text(r"ye").evaluate("e => e.outerHTML")

    page.set_content("<div> ye </div><div>ye</div>")
    assert "> ye </div>" in page.get_by_text("ye", exact=True).first.evaluate(
        "e => e.outerHTML"
    )

    page.set_content("<div>Hello world</div><div>Hello</div>")
    assert (
        page.get_by_text("Hello", exact=True).evaluate("e => e.outerHTML")
        == "<div>Hello</div>"
    )


def test_get_by_label(page: Page) -> None:
    page.set_content(
        "<div><label for=target>Name</label><input id=target type=text></div>"
    )

    expect(page.get_by_label("Name")).to_have_count(1)
    expect(page.main_frame.get_by_label("Name")).to_have_count(1)
    expect(page.locator("div").get_by_label("Name")).to_have_count(1)

    assert page.get_by_text("Name").evaluate("e => e.nodeName") == "LABEL"
    assert page.get_by_label("Name").evaluate("e => e.nodeName") == "INPUT"
    assert page.main_frame.get_by_label("Name").evaluate("e => e.nodeName") == "INPUT"
    assert (
        page.locator("div").get_by_label("Name").evaluate("e => e.nodeName") == "INPUT"
    )


def test_get_by_label_with_nested_elements(page: Page) -> None:
    page.set_content(
        "<label for=target>Last <span>Name</span></label><input id=target type=text>"
    )

    expect(page.get_by_label("last name")).to_have_attribute("id", "target")
    expect(page.get_by_label("st na")).to_have_attribute("id", "target")
    expect(page.get_by_label("Name")).to_have_attribute("id", "target")
    expect(page.get_by_label("Last Name", exact=True)).to_have_attribute("id", "target")
    expect(
        page.get_by_label(re.compile(r"Last\s+name", re.IGNORECASE))
    ).to_have_attribute("id", "target")

    expect(page.get_by_label("Last", exact=True)).to_have_count(0)
    expect(page.get_by_label("last name", exact=True)).to_have_count(0)
    expect(page.get_by_label("Name", exact=True)).to_have_count(0)
    expect(page.get_by_label("what?")).to_have_count(0)
    expect(page.get_by_label(re.compile(r"last name"))).to_have_count(0)


def test_get_by_placeholder(page: Page) -> None:
    page.set_content(
        """<div>
    <input placeholder="Hello">
    <input placeholder="Hello World">
  </div>"""
    )

    expect(page.get_by_placeholder("hello")).to_have_count(2)
    expect(page.main_frame.get_by_placeholder("hello")).to_have_count(2)
    expect(page.locator("div").get_by_placeholder("hello")).to_have_count(2)

    expect(page.get_by_placeholder("hello")).to_have_count(2)
    expect(page.get_by_placeholder("Hello", exact=True)).to_have_count(1)
    expect(page.get_by_placeholder(re.compile(r"wor", re.IGNORECASE))).to_have_count(1)

    # Coverage
    expect(page.main_frame.get_by_placeholder("hello")).to_have_count(2)
    expect(page.locator("div").get_by_placeholder("hello")).to_have_count(2)


def test_get_by_alt_text(page: Page) -> None:
    page.set_content(
        """<div>
    <input alt="Hello">
    <input alt="Hello World">
  </div>"""
    )

    expect(page.get_by_alt_text("hello")).to_have_count(2)
    expect(page.main_frame.get_by_alt_text("hello")).to_have_count(2)
    expect(page.locator("div").get_by_alt_text("hello")).to_have_count(2)

    expect(page.get_by_alt_text("hello")).to_have_count(2)
    expect(page.get_by_alt_text("Hello", exact=True)).to_have_count(1)
    expect(page.get_by_alt_text(re.compile(r"wor", re.IGNORECASE))).to_have_count(1)

    # Coverage
    expect(page.main_frame.get_by_alt_text("hello")).to_have_count(2)
    expect(page.locator("div").get_by_alt_text("hello")).to_have_count(2)


def test_get_by_title(page: Page) -> None:
    page.set_content(
        """<div>
    <input title="Hello">
    <input title="Hello World">
  </div>"""
    )

    expect(page.get_by_title("hello")).to_have_count(2)
    expect(page.main_frame.get_by_title("hello")).to_have_count(2)
    expect(page.locator("div").get_by_title("hello")).to_have_count(2)

    expect(page.get_by_title("hello")).to_have_count(2)
    expect(page.get_by_title("Hello", exact=True)).to_have_count(1)
    expect(page.get_by_title(re.compile(r"wor", re.IGNORECASE))).to_have_count(1)

    # Coverage
    expect(page.main_frame.get_by_title("hello")).to_have_count(2)
    expect(page.locator("div").get_by_title("hello")).to_have_count(2)


def test_get_by_escaping(page: Page) -> None:
    page.set_content(
        """<label id=label for=control>Hello my
wo"rld</label><input id=control />"""
    )
    page.locator("input").evaluate(
        """input => {
    input.setAttribute('placeholder', 'hello my\\nwo"rld');
    input.setAttribute('title', 'hello my\\nwo"rld');
    input.setAttribute('alt', 'hello my\\nwo"rld');
  }"""
    )
    expect(page.get_by_text('hello my\nwo"rld')).to_have_attribute("id", "label")
    expect(page.get_by_text('hello     my         wo"rld')).to_have_attribute(
        "id", "label"
    )
    expect(page.get_by_label('hello my\nwo"rld')).to_have_attribute("id", "control")
    expect(page.get_by_placeholder('hello my\nwo"rld')).to_have_attribute(
        "id", "control"
    )
    expect(page.get_by_alt_text('hello my\nwo"rld')).to_have_attribute("id", "control")
    expect(page.get_by_title('hello my\nwo"rld')).to_have_attribute("id", "control")

    page.set_content(
        """<label id=label for=control>Hello my
world</label><input id=control />"""
    )
    page.locator("input").evaluate(
        """input => {
    input.setAttribute('placeholder', 'hello my\\nworld');
    input.setAttribute('title', 'hello my\\nworld');
    input.setAttribute('alt', 'hello my\\nworld');
  }"""
    )
    expect(page.get_by_text("hello my\nworld")).to_have_attribute("id", "label")
    expect(page.get_by_text("hello      my       world")).to_have_attribute(
        "id", "label"
    )
    expect(page.get_by_label("hello my\nworld")).to_have_attribute("id", "control")
    expect(page.get_by_placeholder("hello my\nworld")).to_have_attribute(
        "id", "control"
    )
    expect(page.get_by_alt_text("hello my\nworld")).to_have_attribute("id", "control")
    expect(page.get_by_title("hello my\nworld")).to_have_attribute("id", "control")


def test_get_by_role(page: Page) -> None:
    page.set_content(
        """
    <button>Hello</button>
    <button>Hel"lo</button>
    <div role="dialog">I am a dialog</div>
  """
    )
    expect(page.get_by_role("button", name="hello")).to_have_count(1)
    expect(page.get_by_role("button", name='Hel"lo')).to_have_count(1)
    expect(
        page.get_by_role("button", name=re.compile(r"he", re.IGNORECASE))
    ).to_have_count(2)
    expect(page.get_by_role("dialog")).to_have_count(1)
